"""
Make an mp4 animation of fgout grid results. 
This is done in a way that makes the animation quickly and with minimum 
storage required, by making one plot and then defining an update function
that only changes the parts of the plot that change in each frame.

Make the animation via:
    python make_fgout_animation.py

If this script is executed in IPython or a notebook it may go into
an infinite loop for reasons unknown.  If so, close the figure to halt.

To view individual fgout frames interactively, this should work:
    import make_fgout_animation
    make_fgout_animation.update(fgframeno)  # for desired fgout frame no

"""

import sys
if 'matplotlib' not in sys.modules:
    # Use an image backend to insure animation has size specified by figsize
    import matplotlib
    matplotlib.use('Agg')

from pylab import *
import os, glob
from clawpack.visclaw import plottools, geoplot
from clawpack.visclaw import animation_tools
from matplotlib import animation, colors
from datetime import timedelta

from clawpack.geoclaw import fgout_tools
    
fgno = 1  # which fgout grid

outdir = '_output'

if 1:
    # use all fgout frames in outdir:
    fgout_frames = glob.glob(os.path.join(outdir, \
                                          'fgout%s.t*' % str(fgno).zfill(4)))
    nout = len(fgout_frames)
    fgframes = range(1, nout+1)
    print('Found %i fgout frames in %s' % (nout,outdir))
else:
    # set explicitly, e.g. to test with only a few frames
    fgframes = range(1,26)  # frames of fgout solution to use in animation

# Instantiate object for reading fgout frames:
fgout_grid = fgout_tools.FGoutGrid(fgno, outdir) 
fgout_grid.read_fgout_grids_data()


# Plot one frame of fgout data and define the Artists that will need to
# be updated in subsequent frames:

fgout1 = fgout_grid.read_frame(fgframes[0])

plot_extent = fgout1.extent_edges
ylat = fgout1.Y.mean()  # for aspect ratio of plots

fig,ax = subplots(figsize=(8,7))

ax.set_xlim(plot_extent[:2])
ax.set_ylim(plot_extent[2:])

plottools.pcolorcells(fgout1.X,fgout1.Y,fgout1.B, cmap=geoplot.land_colors)
clim(0,100)

eta = ma.masked_where(fgout1.h<0.001, fgout1.eta)

eta_plot = plottools.pcolorcells(fgout1.X,fgout1.Y,eta,
                                 cmap=geoplot.tsunami_colormap)
clim(-0.3,0.3)
cb = colorbar(eta_plot, extend='both', shrink=0.7)
cb.set_label('meters')
title_text = title('Surface at time %s' % timedelta(seconds=fgout1.t))

ax.set_aspect(1./cos(ylat*pi/180.))
ticklabel_format(useOffset=False)
xticks(rotation=20)
ax.set_xlim(plot_extent[:2])
ax.set_ylim(plot_extent[2:])


blit = False
if blit:
    # The artists that will be updated for subsequent frames:
    update_artists = (eta_plot, title_text)
    
# Note that update_artists is only needed if blit==True in call to
# animation.FuncAnimation below.
# Using blit==False does not seem to slow down creation of the animation by much
# and slightly simplifies modification of this script to situations where more
# artists are updated.
        
def update(fgframeno):
    """
    Update an exisiting plot with solution from fgout frame fgframeno.
    Note: Even if blit==True in call to animation.FuncAnimation,
    the update_artists do not need to be passed in, unpacked, and repacked
    as in an earlier version of this example (Clawpack version <= 5.10.0).
    """
    
    fgout = fgout_grid.read_frame(fgframeno)
    print('Updating plot at time %s' % timedelta(seconds=fgout.t))
        
    # reset title to current time:
    title_text.set_text('Surface at time %s' % timedelta(seconds=fgout.t))

    # reset surface eta to current state:
    eta = ma.masked_where(fgout.h<0.001, fgout.eta)
    eta_plot.set_array(eta.T.flatten())
        
    if blit:
        return update_artists


if __name__ == '__main__':

    print('Making anim...')
    anim = animation.FuncAnimation(fig, update, frames=fgframes, 
                                   interval=200, blit=blit)
    
    # Output files:
    name = 'fgout_animation'

    fname_mp4 = name + '.mp4'

    #fname_html = None
    fname_html = name + '.html'

    if fname_mp4:
        fps = 5
        print('Making mp4...')
        writer = animation.writers['ffmpeg'](fps=fps)
        anim.save(fname_mp4, writer=writer)
        print("Created %s" % fname_mp4)

    if fname_html:
        # html version:
        animation_tools.make_html(anim, file_name=fname_html, title=name)




r"""Generate C code to compute a 2k-1 order WENO reconstructions.

This generates a Python extension module (written in C) to perform
WENO reconstructions.

Usage:

$ python codegen.py
$ python setup.py build
$ cp build/lib*/reconstruct.so .

Note: the naming convection in PyWENO is cell based: 'left' means the
left edge of a cell, and 'right' means the right edge of a cell.
Matching this up with how PyClaw expects 'ql' and 'qr' to be indexed
is a bit tricky.

"""

import os

import pyweno.symbolic
import pyweno.c

# config

K = list(range(3, 10))                        # 2*k-1 order reconstructions
module = 'reconstruct'                  # py module name
output = module + '.c'

# open output and init

f = open(output, 'w')

c = pyweno.c.CCodeGenerator()
f.write(c.wrapper_head(module))

# smoothness
for k in K:

    print('generating code for k = %d...' % k)

    beta = pyweno.symbolic.jiang_shu_smoothness_coefficients(k)
    c.set_smoothness(beta)

    f.write(c.uniform_smoothness(function='smoothness_k'+str(k),
                                 wrapper=True))

    # left edge of cell (right side of boundary)

    (varpi, split) = pyweno.symbolic.optimal_weights(k, 'left')
    coeffs = pyweno.symbolic.reconstruction_coefficients(k, 'left')

    c.set_optimal_weights(varpi, split)
    c.set_reconstruction_coefficients(coeffs)

    f.write(c.uniform_weights(function='weights_left_k' + str(k),
                              wrapper=True))
    f.write(c.uniform_reconstruction(function='reconstruct_left_k' + str(k),
                                     wrapper=True))

    # right edge of cell (left side boundary, shifted)

    (varpi, split) = pyweno.symbolic.optimal_weights(k, 'right')
    coeffs = pyweno.symbolic.reconstruction_coefficients(k, 'right')

    c.set_optimal_weights(varpi, split)
    c.set_reconstruction_coefficients(coeffs)

    f.write(c.uniform_weights(function='weights_right_k' + str(k),
                              wrapper=True))
    f.write(c.uniform_reconstruction(function='reconstruct_right_k' + str(k),
                                     wrapper=True))


f.write(c.wrapper_foot())
f.close()

try:
    import os
    os.system('indent ' + output)
except:
    pass





import os, sys

from clawpack.visclaw import geoplot
import numpy
from scipy import fft
from numpy import exp,cos,pi
from pylab import figure,clf,plot,title

xlimits = [0, 5e3]

#outdir2 = '_output_100m_sgn'  # second outdir for comparison plots, if desired
outdir2 = None


def setplot(plotdata):

    plotdata.clearfigures()

    def set_dry_tol(current_data):
        current_data.user['dry_tolerance'] = 1e-6

    plotdata.beforeframe = set_dry_tol

    plotfigure = plotdata.new_plotfigure(name='domain', figno=0)
    plotfigure.kwargs = {'figsize':(8,7)}
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.axescmd = 'subplot(211)'
    plotaxes.xlimits = xlimits
    #plotaxes.ylimits = [-5,15]
    plotaxes.title = 'Surface displacement'
    plotaxes.grid = True
    
    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.plot_var = geoplot.surface
    plotitem.color = 'b'

    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.show = False
    plotitem.plot_var = geoplot.topo
    plotitem.color = 'g'

    if outdir2 is not None:
        plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
        plotitem.outdir = outdir2
        plotitem.plot_var = geoplot.surface
        plotitem.color = 'r'
        #plotitem.plotstyle = '-+'


    plotaxes = plotfigure.new_plotaxes()
    plotaxes.axescmd = 'subplot(212)'
    plotaxes.xlimits = xlimits
    #plotaxes.ylimits = [-1,1]
    plotaxes.title = 'Velocity'
    plotaxes.grid = True
        

    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.plot_var = geoplot.u_velocity
    plotitem.color = 'b'

    if outdir2 is not None:
        plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
        plotitem.outdir = outdir2
        plotitem.plot_var = geoplot.u_velocity
        plotitem.color = 'r'



    #-----------------------------------------
    # Figures for gauges
    #-----------------------------------------
    plotfigure = plotdata.new_plotfigure(name='q', figno=300, \
                                         type='each_gauge')
    plotfigure.kwargs = {'figsize':(12,3)}
    plotfigure.clf_each_gauge = True

    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = 'auto'
    #plotaxes.ylimits = [-2,2]
    plotaxes.title = 'Surface elevation eta'

    def addgrid(current_data):
        from pylab import grid, xlim
        grid(True)
        gno = current_data.gaugeno
        if gno == 150:
            xlim(0,60)
        elif gno == 80:
            xlim(10,60)
        elif gno == 60:
            xlim(20,70)
        elif gno == 40:
            xlim(40,90)
        elif gno == 30:
            xlim(40,90)
        elif gno == 20:
            xlim(50,100)
        elif gno == 10:
            xlim(60,110)
    plotaxes.afteraxes = addgrid

    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.plot_var = 2  # eta
    plotitem.plotstyle = 'b-'


    plotdata.printfigs = True          # Whether to output figures
    plotdata.print_format = 'png'      # What type of output format
    plotdata.print_framenos = 'all'      # Which frames to output
    plotdata.print_fignos = 'all'      # Which figures to print
    plotdata.html = True               # Whether to create HTML files
    plotdata.latex = False             # Whether to make LaTeX output
    plotdata.parallel = True

    return plotdata



#!/usr/bin/env python
# encoding: utf-8
r"""
Euler 2D Quadrants example
==========================

Simple example solving the Euler equations of compressible fluid dynamics:

.. math::
    \rho_t + (\rho u)_x + (\rho v)_y & = 0 \\
    (\rho u)_t + (\rho u^2 + p)_x + (\rho uv)_y & = 0 \\
    (\rho v)_t + (\rho uv)_x + (\rho v^2 + p)_y & = 0 \\
    E_t + (u (E + p) )_x + (v (E + p))_y & = 0.

Here :math:`\rho` is the density, (u,v) is the velocity, and E is the total energy.
The initial condition is one of the 2D Riemann problems from the paper of
Liska and Wendroff.

"""
from clawpack import riemann
from clawpack.riemann.euler_4wave_2D_constants import density, x_momentum, y_momentum, \
        energy, num_eqn
from clawpack.visclaw import colormaps

def setplot(plotdata):
    r"""Plotting settings

    Should plot two figures both of density.

    """


    plotdata.clearfigures()  # clear any old figures,axes,items data

    # Figure for density - pcolor
    plotfigure = plotdata.new_plotfigure(name='Density', figno=0)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = 'auto'
    plotaxes.ylimits = 'auto'
    plotaxes.scaled = True
    plotaxes.title = 'Density'

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
    plotitem.plot_var = density
    plotitem.pcolor_cmap = colormaps.yellow_red_blue
    plotitem.pcolor_cmin = 0.
    plotitem.pcolor_cmax = 2.
    plotitem.add_colorbar = True

    # Figure for density - Schlieren
    plotfigure = plotdata.new_plotfigure(name='Schlieren', figno=1)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = 'auto'
    plotaxes.ylimits = 'auto'
    plotaxes.title = 'Density'
    plotaxes.scaled = True      # so aspect ratio is 1

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='2d_schlieren')
    plotitem.schlieren_cmin = 0.0
    plotitem.schlieren_cmax = 1.0
    plotitem.plot_var = density
    plotitem.add_colorbar = False
    
    return plotdata


def setup(use_petsc=False,riemann_solver='roe'):
    if use_petsc:
        import clawpack.petclaw as pyclaw
    else:
        from clawpack import pyclaw

    if riemann_solver.lower() == 'roe':
        solver = pyclaw.ClawSolver2D(riemann.euler_4wave_2D)
        solver.transverse_waves = 2
    elif riemann_solver.lower() == 'hlle':
        solver = pyclaw.ClawSolver2D(riemann.euler_hlle_2D)
        solver.transverse_waves = 0
        solver.cfl_desired = 0.4
        solver.cfl_max = 0.5
    solver.all_bcs = pyclaw.BC.extrap

    domain = pyclaw.Domain([0.,0.],[1.,1.],[100,100])
    solution = pyclaw.Solution(num_eqn,domain)
    gamma = 1.4
    solution.problem_data['gamma']  = gamma

    # Set initial data
    xx, yy = domain.grid.p_centers
    l = xx < 0.8
    r = xx >= 0.8
    b = yy < 0.8
    t = yy >= 0.8
    solution.q[density,...] = 1.5 * r * t + 0.532258064516129 * l * t          \
                                          + 0.137992831541219 * l * b          \
                                          + 0.532258064516129 * r * b
    u = 0.0 * r * t + 1.206045378311055 * l * t                                \
                    + 1.206045378311055 * l * b                                \
                    + 0.0 * r * b
    v = 0.0 * r * t + 0.0 * l * t                                              \
                    + 1.206045378311055 * l * b                                \
                    + 1.206045378311055 * r * b
    p = 1.5 * r * t + 0.3 * l * t + 0.029032258064516 * l * b + 0.3 * r * b
    solution.q[x_momentum,...] = solution.q[density, ...] * u
    solution.q[y_momentum,...] = solution.q[density, ...] * v
    solution.q[energy,...] = 0.5 * solution.q[density,...]*(u**2 + v**2) + p / (gamma - 1.0)

    claw = pyclaw.Controller()
    claw.tfinal = 0.8
    claw.num_output_times = 10
    claw.solution = solution
    claw.solver = solver

    claw.output_format = 'ascii'    
    claw.outdir = "./_output"
    claw.keep_copy = True
    claw.setplot = setplot

    return claw

if __name__ == "__main__":
    from clawpack.pyclaw.util import run_app_from_main
    output = run_app_from_main(setup, setplot)




""" 
Set up the plot figures, axes, and items to be done for each frame.

This module is imported by the plotting routines and then the
function setplot is called to set the plot parameters.
    
""" 

from clawpack.geoclaw import topotools
from clawpack.geoclaw.data import Rearth  # radius of earth
from clawpack.clawutil.data import ClawData

from mapper import latlong, gcdist

probdata = ClawData()
probdata.read('setprob.data', force=True)
theta_island = probdata.theta_island


#--------------------------
def setplot(plotdata):
#--------------------------
    
    """ 
    Specify what is to be plotted at each frame.
    Input:  plotdata, an instance of clawpack.visclaw.data.ClawPlotData.
    Output: a modified version of plotdata.
    
    """ 


    from clawpack.visclaw import colormaps, geoplot
    from numpy import linspace

    plotdata.clearfigures()  # clear any old figures,axes,items data
    plotdata.format = 'binary'


    # To plot gauge locations on pcolor or contour plot, use this as
    # an afteraxis function:

    def addgauges(current_data):
        from clawpack.visclaw import gaugetools
        gaugetools.plot_gauge_locations(current_data.plotdata, \
             gaugenos='all', format_string='ko', add_labels=True)
    

    #-----------------------------------------
    # Figure for pcolor plot
    #-----------------------------------------
    plotfigure = plotdata.new_plotfigure(name='pcolor', figno=0)

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes('pcolor')
    plotaxes.title = 'Surface'
    plotaxes.scaled = True

    def fixup(current_data):
        import pylab
        t = current_data.t
        #addgauges(current_data)
        pylab.title("Surface at t = %8.1f" % t, fontsize=20)
        #pylab.xticks(fontsize=15)
        #pylab.yticks(fontsize=15)
        pylab.gca().set_aspect(1./pylab.cos(40*pylab.pi/180.))
    plotaxes.afteraxes = fixup

    # Water
    plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
    #plotitem.plot_var = geoplot.surface
    plotitem.plot_var = geoplot.surface_or_depth
    plotitem.pcolor_cmap = geoplot.tsunami_colormap
    plotitem.pcolor_cmin = -1.0
    plotitem.pcolor_cmax = 1.0
    plotitem.add_colorbar = True
    plotitem.amr_celledges_show = [0,0,0]
    plotitem.patchedges_show = 1

    # Land
    plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
    plotitem.plot_var = geoplot.land
    #plotitem.pcolor_cmap = colormaps.all_white  # to print better in B&W
    plotitem.pcolor_cmap = geoplot.land_colors
    plotitem.pcolor_cmin = 0.0
    plotitem.pcolor_cmax = 100.0
    plotitem.add_colorbar = False
    plotitem.amr_celledges_show = [1,0]
    plotitem.patchedges_show = 1
    plotaxes.xlimits = [-20,20]
    plotaxes.ylimits = [20,60]


    # add contour lines of bathy if desired:
    plotitem = plotaxes.new_plotitem(plot_type='2d_contour')
    plotitem.show = False
    plotitem.plot_var = geoplot.topo
    plotitem.contour_levels = linspace(-1000,-1000,1)
    plotitem.amr_contour_colors = ['g']  # color on each level
    plotitem.kwargs = {'linestyles':'dashed','linewidths':2}
    plotitem.amr_contour_show = [1,0,0]  
    plotitem.celledges_show = 0
    plotitem.patchedges_show = 0


    #-----------------------------------------
    # Figure for zoom
    #-----------------------------------------
    plotfigure = plotdata.new_plotfigure(name='Zoom1', figno=7)
    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes('zoom on island')
    plotaxes.title = 'Surface elevation'
    xisland,yisland = latlong(1600e3, theta_island, 40., Rearth)
    plotaxes.xlimits = [xisland-0.6, xisland+0.8]
    plotaxes.ylimits = [yisland-0.6, yisland+0.8]
    def fixup(current_data):
        import pylab
        t = current_data.t
        addgauges(current_data)
        pylab.title("Surface at t = %8.1f" % t, fontsize=20)
        #pylab.xticks(fontsize=15)
        #pylab.yticks(fontsize=15)
        pylab.gca().set_aspect(1./pylab.cos(50*pylab.pi/180.))
    plotaxes.afteraxes = fixup

    # Water
    plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
    plotitem.show = True
    #plotitem.plot_var = geoplot.surface
    plotitem.plot_var = geoplot.surface_or_depth
    plotitem.pcolor_cmap = geoplot.tsunami_colormap
    plotitem.pcolor_cmin = -1.0
    plotitem.pcolor_cmax = 1.0
    plotitem.add_colorbar = False
    plotitem.amr_celledges_show = [0]
    plotitem.patchedges_show = 0

    # Land
    plotitem = plotaxes.new_plotitem(plot_type='2d_pcolor')
    plotitem.show = True
    plotitem.plot_var = geoplot.land
    #plotitem.pcolor_cmap = colormaps.all_white  # to print better in B&W
    plotitem.pcolor_cmap = geoplot.land_colors
    plotitem.pcolor_cmin = 0.0
    plotitem.pcolor_cmax = 100.0
    plotitem.add_colorbar = False
    plotitem.amr_celledges_show = [1,0,0,0,0]
    plotitem.patchedges_show = 1

    # contour lines:
    plotitem = plotaxes.new_plotitem(plot_type='2d_contour')
    plotitem.show = True
    plotitem.plot_var = geoplot.surface
    plotitem.contour_levels = [-0.8, -0.4, 0.4, 0.8]
    plotitem.amr_contour_colors = ['k']  # color on each level
    plotitem.kwargs = {'linewidths':2}
    plotitem.amr_contour_show = [0,0,0,1,1]  
    plotitem.celledges_show = 0
    plotitem.patchedges_show = 0

    # add contour lines of bathy if desired:
    plotitem = plotaxes.new_plotitem(plot_type='2d_contour')
    plotitem.show = False
    plotitem.plot_var = geoplot.topo
    plotitem.contour_levels = linspace(-1000,-1000,1)
    plotitem.amr_contour_colors = ['g']  # color on each level
    plotitem.kwargs = {'linestyles':'dashed','linewidths':2}
    plotitem.amr_contour_show = [0,0,1]  
    plotitem.celledges_show = 0
    plotitem.patchedges_show = 0

    #-----------------------------------------
    # Figures for gauges
    #-----------------------------------------
    plotfigure = plotdata.new_plotfigure(name='Surface & topo', figno=300, \
                    type='each_gauge')
    plotfigure.clf_each_gauge = True

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = [6000, 14000]
    plotaxes.ylimits = [-2.5, 4.0]
    plotaxes.title = 'Surface'

    # Plot surface as blue curve:
    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.plot_var = 3
    plotitem.plotstyle = 'b-'

    # Plot topo as green curve:
    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')

    def gaugetopo(current_data):
        q = current_data.q
        h = q[0,:]
        eta = q[3,:]
        topo = eta - h
        return topo
        
    plotitem.plot_var = gaugetopo
    plotitem.plotstyle = 'g-'
    def add_zeroline(current_data):
        from pylab import plot, legend
        t = current_data.t
        #legend(('surface','topography'),loc='lower left')
        plot(t, 0*t, 'k')

    plotaxes.afteraxes = add_zeroline


    #-----------------------------------------
    # Figure for grids alone
    #-----------------------------------------
    plotfigure = plotdata.new_plotfigure(name='grids', figno=2)
    plotfigure.show = False

    # Set up for axes in this figure:
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = [0,1]
    plotaxes.ylimits = [0,1]
    plotaxes.title = 'grids'
    plotaxes.scaled = True

    # Set up for item on these axes:
    plotitem = plotaxes.new_plotitem(plot_type='2d_patch')
    plotitem.amr_patch_bgcolor = ['#ffeeee', '#eeeeff', '#eeffee']
    plotitem.amr_celledges_show = [1,1,0]   
    plotitem.amr_patchedges_show = [1]     



    #-----------------------------------------
    # Figures for fgmax plots
    #-----------------------------------------
    # Note: You need to move fgmax png files into _plots/fgmax_plots after
    # creating them, e.g., by running the process_fgmax notebook or script.
    # The lines below just create links to these figures from _PlotIndex.html

    otherfigure = plotdata.new_otherfigure(name='max depth on fgmax grid 1',
                    fname='fgmax_plots/fgmax0001_h_onshore.png')

    otherfigure = plotdata.new_otherfigure(name='max speed on fgmax grid 1',
                    fname='fgmax_plots/fgmax0001_speed.png')

    otherfigure = plotdata.new_otherfigure(name='max elevation on fgmax grid 2',
                    fname='fgmax_plots/fgmax0002_surface.png')

    otherfigure = plotdata.new_otherfigure(name='max depth on fgmax grid 3',
                    fname='fgmax_plots/fgmax0003_h_onshore.png')

    otherfigure = plotdata.new_otherfigure(name='max speed on fgmax grid 3',
                    fname='fgmax_plots/fgmax0003_speed.png')

    # add additional lines for any other figures you want added to the index.    

    #-----------------------------------------
    # Plots of timing (CPU and wall time):

    def make_timing_plots(plotdata):
        import os
        from clawpack.visclaw import plot_timing_stats
        try:
            timing_plotdir = plotdata.plotdir + '/timing_figures'
            os.system('mkdir -p %s' % timing_plotdir)
            units = {'comptime':'seconds', 'simtime':'hours', 'cell':'millions'}
            plot_timing_stats.make_plots(outdir=plotdata.outdir, make_pngs=True,
                                          plotdir=timing_plotdir, units=units)
            os.system('cp %s/timing.* %s' % (plotdata.outdir, timing_plotdir))
        except:
            print('*** Error making timing plots')

    # create a link to this webpage from _PlotIndex.html:
    otherfigure = plotdata.new_otherfigure(name='timing',
                    fname='timing_figures/timing.html')
    otherfigure.makefig = make_timing_plots



    #-----------------------------------------
    
    # Parameters used only when creating html and/or latex hardcopy
    # e.g., via clawpack.visclaw.frametools.printframes:

    plotdata.printfigs = True                # print figures
    plotdata.print_format = 'png'            # file format
    plotdata.print_framenos = 'all'          # list of frames to print
    plotdata.print_gaugenos = 'all'          # list of gauges to print
    plotdata.print_fignos = 'all'            # list of figures to print
    plotdata.html = True                     # create html files of plots?
    plotdata.html_homelink = '../README.html'   # pointer for top of index
    plotdata.latex = True                    # create latex file of plots?
    plotdata.latex_figsperline = 2           # layout of plots
    plotdata.latex_framesperline = 1         # layout of plots
    plotdata.latex_makepdf = False           # also run pdflatex?

    return plotdata

    



r"""
Module containing the classic Clawpack solvers.

This module contains the pure and wrapped classic clawpack solvers.  All 
clawpack solvers inherit from the :class:`ClawSolver` superclass which in turn 
inherits from the :class:`~pyclaw.solver.Solver` superclass.  These
are both pure virtual classes; the only solver classes that should be instantiated
are the dimension-specific ones, :class:`ClawSolver1D` and :class:`ClawSolver2D`.
"""

from clawpack.pyclaw.util import add_parent_doc
from clawpack.pyclaw.solver import Solver
from clawpack.pyclaw.limiters import tvd
from importlib import reload

# ============================================================================
#  Generic Clawpack solver class
# ============================================================================
class ClawSolver(Solver):
    r"""
    Generic classic Clawpack solver
    
    All Clawpack solvers inherit from this base class.
    
    .. attribute:: mthlim 
    
        Limiter(s) to be used.  Specified either as one value or a list.
        If one value, the specified limiter is used for all wave families.
        If a list, the specified values indicate which limiter to apply to
        each wave family.  Take a look at pyclaw.limiters.tvd for an enumeration.
        ``Default = limiters.tvd.minmod``
    
    .. attribute:: order
    
        Order of the solver, either 1 for first order (i.e., Godunov's method)
        or 2 for second order (Lax-Wendroff-LeVeque).
        ``Default = 2``
    
    .. attribute:: source_split
    
        Which source splitting method to use: 1 for first 
        order Godunov splitting and 2 for second order Strang splitting.
        ``Default = 1``
        
    .. attribute:: fwave
    
        Whether to split the flux jump (rather than the jump in Q) into waves; 
        requires that the Riemann solver performs the splitting.  
        ``Default = False``
        
    .. attribute:: step_source
    
        Handle for function that evaluates the source term.  
        The required signature for this function is:

        def step_source(solver,state,dt)

    .. attribute:: kernel_language

        Specifies whether to use wrapped Fortran routines ('Fortran')
        or pure Python ('Python').  ``Default = 'Fortran'``.
    
    .. attribute:: verbosity

        The level of detail of logged messages from the Fortran solver.
        ``Default = 0``.

    """
    
    # ========== Generic Init Routine ========================================
    def __init__(self,riemann_solver=None,claw_package=None):
        r"""
        See :class:`ClawSolver` for full documentation.

        Output:
        - (:class:`ClawSolver`) - Initialized clawpack solver
        """
        self.num_ghost = 2
        self.limiters = tvd.minmod
        self.order = 2
        self.source_split = 1
        self.fwave = False
        self.step_source = None
        self.kernel_language = 'Fortran'
        self.verbosity = 0
        self.cfl_max = 1.0
        self.cfl_desired = 0.9
        self._mthlim = self.limiters
        self._method = None
        self.dt_old = None

        # Call general initialization function
        super(ClawSolver,self).__init__(riemann_solver,claw_package)
    
    # ========== Time stepping routines ======================================
    def step(self,solution,take_one_step,tstart,tend):
        r"""
        Evolve solution one time step

        The elements of the algorithm for taking one step are:

        1. Pick a step size as specified by the base solver attribute :func:`get_dt`
        
        2. A half step on the source term :func:`step_source` if Strang splitting is 
           being used (:attr:`source_split` = 2)
        
        3. A step on the homogeneous problem :math:`q_t + f(q)_x = 0` is taken
        
        4. A second half step or a full step is taken on the source term
           :func:`step_source` depending on whether Strang splitting was used 
           (:attr:`source_split` = 2) or Godunov splitting 
           (:attr:`source_split` = 1)

        This routine is called from the method evolve_to_time defined in the
        pyclaw.solver.Solver superclass.

        :Input:
         - *solution* - (:class:`~pyclaw.solution.Solution`) solution to be evolved
         
        :Output: 
         - (bool) - True if full step succeeded, False otherwise
        """
        self.get_dt(solution.t,tstart,tend,take_one_step)
        self.cfl.set_global_max(0.)

        if self.source_split == 2 and self.step_source is not None:
            self.step_source(self,solution.states[0],self.dt/2.0)

        self.step_hyperbolic(solution)

        # Check here if the CFL condition is satisfied. 
        # If not, return # immediately to evolve_to_time and let it deal with
        # picking a new step size (dt).
        if self.cfl.get_cached_max() >= self.cfl_max:
            return False

        if self.step_source is not None:
            # Strang splitting
            if self.source_split == 2:
                self.step_source(self,solution.states[0],self.dt/2.0)

            # Godunov Splitting
            if self.source_split == 1:
                self.step_source(self,solution.states[0],self.dt)
                
        return True

    def _check_cfl_settings(self):
        pass

    def _allocate_workspace(self,solution):
        pass

    def step_hyperbolic(self,solution):
        r"""
        Take one homogeneous step on the solution.
        
        This is a dummy routine and must be overridden.
        """
        raise Exception("Dummy routine, please override!")

    def _set_mthlim(self):
        r"""
        Convenience routine to convert users limiter specification to 
        the format understood by the Fortran code (i.e., a list of length num_waves).
        """
        self._mthlim = self.limiters
        if not isinstance(self.limiters,list): self._mthlim=[self._mthlim]
        if len(self._mthlim)==1: self._mthlim = self._mthlim * self.num_waves
        if len(self._mthlim)!=self.num_waves:
            raise Exception('Length of solver.limiters is not equal to 1 or to solver.num_waves')
 
    def _set_method(self,state):
        r"""
        Set values of the solver._method array required by the Fortran code.
        These are algorithmic parameters.
        """
        import numpy as np
        #We ought to put method and many other things in a Fortran
        #module and set the fortran variables directly here.
        self._method =np.empty(7, dtype=int,order='F')
        self._method[0] = self.dt_variable
        self._method[1] = self.order
        if self.num_dim==1:
            self._method[2] = 0  # Not used in 1D
        elif self.dimensional_split:
            self._method[2] = -1  # First-order dimensional splitting
        else:
            self._method[2] = self.transverse_waves
        self._method[3] = self.verbosity
        self._method[4] = 0  # Not used for PyClaw (would be self.source_split)
        self._method[5] = state.index_capa + 1
        self._method[6] = state.num_aux

    def setup(self,solution):
        r"""
        Perform essential solver setup.  This routine must be called before
        solver.step() may be called.
        """
        # This is a hack to deal with the fact that petsc4py
        # doesn't allow us to change the stencil_width (num_ghost)
        solution.state.set_num_ghost(self.num_ghost)
        # End hack

        self._check_cfl_settings()

        self._set_mthlim()
        if(self.kernel_language == 'Fortran'):
            if self.fmod is None:
                so_name = 'clawpack.pyclaw.classic.classic'+str(self.num_dim)
                self.fmod = __import__(so_name,fromlist=['clawpack.pyclaw.classic'])
            self._set_fortran_parameters(solution)
            self._allocate_workspace(solution)
        elif self.num_dim>1:
            raise Exception('Only Fortran kernels are supported in multi-D.')

        self._allocate_bc_arrays(solution.states[0])

        super(ClawSolver,self).setup(solution)


    def _set_fortran_parameters(self,solution):
        r"""
        Pack parameters into format recognized by Clawpack (Fortran) code.

        Sets the solver._method array and the cparam common block for the Riemann solver.
        """
        self._set_method(solution.state)
        # The reload here is necessary because otherwise the common block
        # cparam in the Riemann solver doesn't get flushed between running
        # different tests in a single Python session.
        reload(self.fmod)
        solution.state.set_cparam(self.fmod)
        solution.state.set_cparam(self.rp)

    def __del__(self):
        r"""
        Delete Fortran objects, which otherwise tend to persist in Python sessions.
        """
        if(self.kernel_language == 'Fortran'):
            del self.fmod

        super(ClawSolver,self).__del__()


# ============================================================================
#  ClawPack 1d Solver Class
# ============================================================================
class ClawSolver1D(ClawSolver):
    r"""
    Clawpack evolution routine in 1D
    
    This class represents the 1d clawpack solver on a single grid.  Note that 
    there are routines here for interfacing with the fortran time stepping 
    routines and the Python time stepping routines.  The ones used are 
    dependent on the argument given to the initialization of the solver 
    (defaults to python).
    
    """

    __doc__ += add_parent_doc(ClawSolver)

    def __init__(self, riemann_solver=None, claw_package=None):
        r"""
        Create 1d Clawpack solver

        Output:
        - (:class:`ClawSolver1D`) - Initialized 1d clawpack solver
        
        See :class:`ClawSolver1D` for more info.
        """   
        self.num_dim = 1
        self.reflect_index = [1]

        super(ClawSolver1D,self).__init__(riemann_solver, claw_package)


    # ========== Homogeneous Step =====================================
    def step_hyperbolic(self,solution):
        r"""
        Take one time step on the homogeneous hyperbolic system.

        :Input:
         - *solution* - (:class:`~pyclaw.solution.Solution`) Solution that 
           will be evolved
        """
        import numpy as np

        state = solution.states[0]
        grid = state.grid

        self._apply_bcs(state)
            
        num_eqn,num_ghost = state.num_eqn,self.num_ghost
          
        if(self.kernel_language == 'Fortran'):
            mx = grid.num_cells[0]
            dx,dt = grid.delta[0],self.dt
            dtdx = np.zeros( (mx+2*num_ghost) ) + dt/dx
            rp1 = self.rp.rp1._cpointer
            
            self.qbc,cfl = self.fmod.step1(num_ghost,mx,self.qbc,self.auxbc,dx,dt,self._method,self._mthlim,self.fwave,rp1)
            
        elif(self.kernel_language == 'Python'):
 
            q   = self.qbc
            aux = self.auxbc
            # Limiter to use in the pth family
            limiter = np.array(self._mthlim,ndmin=1)  
        
            dtdx = np.zeros( (2*self.num_ghost+grid.num_cells[0]) )

            # Find local value for dt/dx
            if state.index_capa>=0:
                dtdx = self.dt / (grid.delta[0] * aux[state.index_capa,:])
            else:
                dtdx += self.dt/grid.delta[0]
        
            # Solve Riemann problem at each interface
            q_l=q[:,:-1]
            q_r=q[:,1:]
            if state.aux is not None:
                aux_l=aux[:,:-1]
                aux_r=aux[:,1:]
            else:
                aux_l = None
                aux_r = None
            wave,s,amdq,apdq = self.rp(q_l,q_r,aux_l,aux_r,state.problem_data)
            
            # Update loop limits, these are the limits for the Riemann solver
            # locations, which then update a grid cell value
            # We include the Riemann problem just outside of the grid so we can
            # do proper limiting at the grid edges
            #        LL    |                               |     UL
            #  |  LL |     |     |     |  ...  |     |     |  UL  |     |
            #              |                               |

            LL = self.num_ghost - 1
            UL = self.num_ghost + grid.num_cells[0] + 1 

            # Update q for Godunov update
            for m in range(num_eqn):
                q[m,LL:UL] -= dtdx[LL:UL]*apdq[m,LL-1:UL-1]
                q[m,LL-1:UL-1] -= dtdx[LL-1:UL-1]*amdq[m,LL-1:UL-1]
        
            # Compute maximum wave speed
            cfl = 0.0
            for mw in range(wave.shape[1]):
                smax1 = np.max(dtdx[LL:UL]*s[mw,LL-1:UL-1])
                smax2 = np.max(-dtdx[LL-1:UL-1]*s[mw,LL-1:UL-1])
                cfl = max(cfl,smax1,smax2)

            # If we are doing slope limiting we have more work to do
            if self.order == 2:
                # Initialize flux corrections
                f = np.zeros( (num_eqn,grid.num_cells[0] + 2*self.num_ghost) )
            
                # Apply Limiters to waves
                if (limiter > 0).any():
                    wave = tvd.limit(state.num_eqn,wave,s,limiter,dtdx)

                # Compute correction fluxes for second order q_{xx} terms
                dtdxave = 0.5 * (dtdx[LL-1:UL-1] + dtdx[LL:UL])
                if self.fwave:
                    for mw in range(wave.shape[1]):
                        sabs = np.abs(s[mw,LL-1:UL-1])
                        om = 1.0 - sabs*dtdxave[:UL-LL]
                        ssign = np.sign(s[mw,LL-1:UL-1])
                        for m in range(num_eqn):
                            f[m,LL:UL] += 0.5 * ssign * om * wave[m,mw,LL-1:UL-1]
                else:
                    for mw in range(wave.shape[1]):
                        sabs = np.abs(s[mw,LL-1:UL-1])
                        om = 1.0 - sabs*dtdxave[:UL-LL]
                        for m in range(num_eqn):
                            f[m,LL:UL] += 0.5 * sabs * om * wave[m,mw,LL-1:UL-1]

                # Update q by differencing correction fluxes
                for m in range(num_eqn):
                    q[m,LL:UL-1] -= dtdx[LL:UL-1] * (f[m,LL+1:UL] - f[m,LL:UL-1]) 

        else: raise Exception("Unrecognized kernel_language; choose 'Fortran' or 'Python'")

        self.cfl.update_global_max(cfl)
        state.set_q_from_qbc(num_ghost,self.qbc)
        if state.num_aux > 0:
            state.set_aux_from_auxbc(num_ghost,self.auxbc)
   

# ============================================================================
#  ClawPack 2d Solver Class
# ============================================================================
class ClawSolver2D(ClawSolver):
    r"""
    2D Classic (Clawpack) solver.

    Solve using the wave propagation algorithms of Randy LeVeque's
    Clawpack code (www.clawpack.org).

    In addition to the attributes of ClawSolver1D, ClawSolver2D
    also has the following options:
    
    .. attribute:: dimensional_split
    
        If True, use dimensional splitting (Godunov splitting).
        Dimensional splitting with Strang splitting is not supported
        at present but could easily be enabled if necessary.
        If False, use unsplit Clawpack algorithms, possibly including
        transverse Riemann solves.

    .. attribute:: transverse_waves
    
        If dimensional_split is True, this option has no effect.  If
        dimensional_split is False, then transverse_waves should be one of
        the following values:

        ClawSolver2D.no_trans: Transverse Riemann solver
        not used.  The stable CFL for this algorithm is 0.5.  Not recommended.
        
        ClawSolver2D.trans_inc: Transverse increment waves are computed
        and propagated.

        ClawSolver2D.trans_cor: Transverse increment waves and transverse
        correction waves are computed and propagated.

    Note that only the fortran routines are supported for now in 2D.
    """

    __doc__ += add_parent_doc(ClawSolver)
    
    no_trans  = 0
    trans_inc = 1
    trans_cor = 2

    def __init__(self,riemann_solver=None, claw_package=None):
        r"""
        Create 2d Clawpack solver
        
        See :class:`ClawSolver2D` for more info.
        """   
        self.dimensional_split = True
        self.transverse_waves = self.trans_inc

        self.num_dim = 2
        self.reflect_index = [1,2]

        self.aux1 = None
        self.aux2 = None
        self.aux3 = None
        self.work = None

        super(ClawSolver2D,self).__init__(riemann_solver, claw_package)

    def _check_cfl_settings(self):
        if (not self.dimensional_split) and (self.transverse_waves==0):
            cfl_recommended = 0.5
        else:
            cfl_recommended = 1.0

        if self.cfl_max > cfl_recommended:
            import warnings
            warnings.warn('cfl_max is set higher than the recommended value of %s' % cfl_recommended)
            warnings.warn(str(self.cfl_desired))


    def _allocate_workspace(self,solution):
        r"""
        Pack parameters into format recognized by Clawpack (Fortran) code.

        Sets the method array and the cparam common block for the Riemann solver.
        """
        import numpy as np

        state = solution.state

        num_eqn,num_aux,num_waves,num_ghost,aux = state.num_eqn,state.num_aux,self.num_waves,self.num_ghost,state.aux

        #The following is a hack to work around an issue
        #with f2py.  It involves wastefully allocating three arrays.
        #f2py seems not able to handle multiple zero-size arrays being passed.
        # it appears the bug is related to f2py/src/fortranobject.c line 841.
        if aux is None: num_aux=1

        grid  = state.grid
        maxmx,maxmy = grid.num_cells[0],grid.num_cells[1]
        maxm = max(maxmx, maxmy)

        # These work arrays really ought to live inside a fortran module
        # as is done for sharpclaw
        self.aux1 = np.empty((num_aux,maxm+2*num_ghost),order='F')
        self.aux2 = np.empty((num_aux,maxm+2*num_ghost),order='F')
        self.aux3 = np.empty((num_aux,maxm+2*num_ghost),order='F')
        mwork = (maxm+2*num_ghost) * (5*num_eqn + num_waves + num_eqn*num_waves)
        self.work = np.empty((mwork),order='F')


    # ========== Hyperbolic Step =====================================
    def step_hyperbolic(self,solution):
        r"""
        Take a step on the homogeneous hyperbolic system using the Clawpack
        algorithm.

        Clawpack is based on the Lax-Wendroff method, combined with Riemann
        solvers and TVD limiters applied to waves.
        """
        if(self.kernel_language == 'Fortran'):
            state = solution.states[0]
            grid = state.grid
            dx,dy = grid.delta
            mx,my = grid.num_cells
            maxm = max(mx,my)
            
            self._apply_bcs(state)
            qold = self.qbc.copy('F')
            
            rpn2 = self.rp.rpn2._cpointer

            if (self.dimensional_split) or (self.transverse_waves==0):
                rpt2 = rpn2 # dummy value; it won't be called
            else:
                rpt2 = self.rp.rpt2._cpointer

            if self.dimensional_split:
                #Right now only Godunov-dimensional-splitting is implemented.
                #Strang-dimensional-splitting could be added following dimsp2.f in Clawpack.

                self.qbc, cfl_x = self.fmod.step2ds(maxm,self.num_ghost,mx,my, \
                      qold,self.qbc,self.auxbc,dx,dy,self.dt,self._method,self._mthlim,\
                      self.aux1,self.aux2,self.aux3,self.work,1,self.fwave,rpn2,rpt2)

                self.qbc, cfl_y = self.fmod.step2ds(maxm,self.num_ghost,mx,my, \
                      self.qbc,self.qbc,self.auxbc,dx,dy,self.dt,self._method,self._mthlim,\
                      self.aux1,self.aux2,self.aux3,self.work,2,self.fwave,rpn2,rpt2)

                cfl = max(cfl_x,cfl_y)

            else:

                self.qbc, cfl = self.fmod.step2(maxm,self.num_ghost,mx,my, \
                      qold,self.qbc,self.auxbc,dx,dy,self.dt,self._method,self._mthlim,\
                      self.aux1,self.aux2,self.aux3,self.work,self.fwave,rpn2,rpt2)

            self.cfl.update_global_max(cfl)
            state.set_q_from_qbc(self.num_ghost,self.qbc)
            if state.num_aux > 0:
                state.set_aux_from_auxbc(self.num_ghost,self.auxbc)

        else:
            raise NotImplementedError("No python implementation for step_hyperbolic in 2D.")

# ============================================================================
#  ClawPack 3d Solver Class
# ============================================================================
class ClawSolver3D(ClawSolver):
    r"""
    3D Classic (Clawpack) solver.

    Solve using the wave propagation algorithms of Randy LeVeque's
    Clawpack code (www.clawpack.org).

    In addition to the attributes of ClawSolver, ClawSolver3D
    also has the following options:
    
    .. attribute:: dimensional_split
    
        If True, use dimensional splitting (Godunov splitting).
        Dimensional splitting with Strang splitting is not supported
        at present but could easily be enabled if necessary.
        If False, use unsplit Clawpack algorithms, possibly including
        transverse Riemann solves.

    .. attribute:: transverse_waves
    
        If dimensional_split is True, this option has no effect.  If
        dim_plit is False, then transverse_waves should be one of
        the following values:

        ClawSolver3D.no_trans: Transverse Riemann solver
        not used.  The stable CFL for this algorithm is 0.5.  Not recommended.
        
        ClawSolver3D.trans_inc: Transverse increment waves are computed
        and propagated.

        ClawSolver3D.trans_cor: Transverse increment waves and transverse
        correction waves are computed and propagated.

    Note that only Fortran routines are supported for now in 3D --
    there is no pure-python version.
    """

    __doc__ += add_parent_doc(ClawSolver)

    no_trans  = 0
    trans_inc = 11
    trans_cor = 22

    def __init__(self, riemann_solver=None, claw_package=None):
        r"""
        Create 3d Clawpack solver
        
        See :class:`ClawSolver3D` for more info.
        """   
        # Add the functions as required attributes
        self.dimensional_split = True
        self.transverse_waves = self.trans_cor

        self.num_dim = 3
        self.reflect_index = [1,2,3]

        self.aux1 = None
        self.aux2 = None
        self.aux3 = None
        self.work = None

        super(ClawSolver3D,self).__init__(riemann_solver, claw_package)

    # ========== Setup routine =============================   
    def _allocate_workspace(self,solution):
        r"""
        Allocate auxN and work arrays for use in Fortran subroutines.
        """
        import numpy as np

        state = solution.states[0]

        num_eqn,num_aux,num_waves,num_ghost,aux = state.num_eqn,state.num_aux,self.num_waves,self.num_ghost,state.aux

        #The following is a hack to work around an issue
        #with f2py.  It involves wastefully allocating three arrays.
        #f2py seems not able to handle multiple zero-size arrays being passed.
        # it appears the bug is related to f2py/src/fortranobject.c line 841.
        if(aux is None): num_aux=1

        grid  = state.grid
        maxmx,maxmy,maxmz = grid.num_cells[0],grid.num_cells[1],grid.num_cells[2]
        maxm = max(maxmx, maxmy, maxmz)

        # These work arrays really ought to live inside a fortran module
        # as is done for sharpclaw
        self.aux1 = np.empty((num_aux,maxm+2*num_ghost,3),order='F')
        self.aux2 = np.empty((num_aux,maxm+2*num_ghost,3),order='F')
        self.aux3 = np.empty((num_aux,maxm+2*num_ghost,3),order='F')
        mwork = (maxm+2*num_ghost) * (31*num_eqn + num_waves + num_eqn*num_waves)
        self.work = np.empty((mwork),order='F')


    # ========== Hyperbolic Step =====================================
    def step_hyperbolic(self,solution):
        r"""
        Take a step on the homogeneous hyperbolic system using the Clawpack
        algorithm.

        Clawpack is based on the Lax-Wendroff method, combined with Riemann
        solvers and TVD limiters applied to waves.
        """
        if(self.kernel_language == 'Fortran'):
            state = solution.states[0]
            grid = state.grid
            dx,dy,dz = grid.delta
            mx,my,mz = grid.num_cells
            maxm = max(mx,my,mz)
            
            self._apply_bcs(state)
            qnew = self.qbc
            qold = qnew.copy('F')
            
            rpn3  = self.rp.rpn3._cpointer

            if (self.dimensional_split) or (self.transverse_waves==0):
                rpt3  = rpn3 # dummy value; it won't be called
                rptt3 = rpn3 # dummy value; it won't be called
            else:
                rpt3  = self.rp.rpt3._cpointer
                rptt3 = self.rp.rptt3._cpointer

            if self.dimensional_split:
                #Right now only Godunov-dimensional-splitting is implemented.
                #Strang-dimensional-splitting could be added following dimsp3.f in Clawpack.

                q, cfl_x = self.fmod.step3ds(maxm,self.num_ghost,mx,my,mz, \
                      qold,qnew,self.auxbc,dx,dy,dz,self.dt,self._method,self._mthlim,\
                      self.aux1,self.aux2,self.aux3,self.work,1,self.fwave,rpn3,rpt3,rptt3)

                q, cfl_y = self.fmod.step3ds(maxm,self.num_ghost,mx,my,mz, \
                      q,q,self.auxbc,dx,dy,dz,self.dt,self._method,self._mthlim,\
                      self.aux1,self.aux2,self.aux3,self.work,2,self.fwave,rpn3,rpt3,rptt3)

                q, cfl_z = self.fmod.step3ds(maxm,self.num_ghost,mx,my,mz, \
                      q,q,self.auxbc,dx,dy,dz,self.dt,self._method,self._mthlim,\
                      self.aux1,self.aux2,self.aux3,self.work,3,self.fwave,rpn3,rpt3,rptt3)

                cfl = max(cfl_x,cfl_y,cfl_z)

            else:

                q, cfl = self.fmod.step3(maxm,self.num_ghost,mx,my,mz, \
                      qold,qnew,self.auxbc,dx,dy,dz,self.dt,self._method,self._mthlim,\
                      self.aux1,self.aux2,self.aux3,self.work,self.fwave,rpn3,rpt3,rptt3)

            self.cfl.update_global_max(cfl)
            state.set_q_from_qbc(self.num_ghost,self.qbc)
            if state.num_aux > 0:
                state.set_aux_from_auxbc(self.num_ghost,self.auxbc)

        else:
            raise NotImplementedError("No python implementation for step_hyperbolic in 3D.")



"""
Plotting Data Module

Contains the general class definition and the subclasses of the Clawpack
data objects specific to plotting.
"""
from __future__ import absolute_import
from __future__ import print_function
import os
import copy
import numpy as np
import re
import logging
import clawpack.clawutil.data as clawdata
import time

import clawpack.pyclaw.controller

# ============================================================================
#  Subclass ClawPlotData containing data for plotting results
# ============================================================================
class ClawPlotData(clawdata.ClawData):
    """ClawPlotData class

    Data subclass containing plot data.

    """

    # ========== Initialization routine ======================================
    def __init__(self, controller=None):
        """Initialize a PlotData object

        """

        # Initialize the data object and read the data files
        super(ClawPlotData,self).__init__()

        # default values of attributes:

        if controller:
            controller.plotdata = self
            # inherit some values from controller
            self.add_attribute('rundir',copy.copy(controller.rundir))
            self.add_attribute('outdir',copy.copy(controller.outdir))
            if len(controller.frames)>0:
                for i,frame in enumerate(controller.frames):
                    self.framesoln_dict[str(i)] = frame
            self.add_attribute('format',copy.copy(controller.output_format))
            self.add_attribute('file_prefix',copy.copy(controller.output_file_prefix))
        else:
            self.add_attribute('rundir',os.getcwd())     # uses *.data from rundir
            self.add_attribute('outdir',os.getcwd())     # where to find fort.* files
            self.add_attribute('format','ascii')
            self.add_attribute('file_prefix','fort')

        # This should eventually replace all need for recording the above
        # information
        self.add_attribute('output_controller', None)
        self.output_controller = clawpack.pyclaw.controller.OutputController(
                                           self.outdir, file_format=self.format)


        self.add_attribute('plotdir',os.getcwd())      # directory for plots *.png, *.html
        self.add_attribute('overwrite',True)           # ok to overwrite old plotdir?
        self.add_attribute('plotter','matplotlib')     # backend for plots
        self.add_attribute('msgfile','')               # where to write error messages
        self.add_attribute('verbose',True)             # verbose output?

        self.add_attribute('ion',False)                # call ion() or ioff()?

        self.add_attribute('printfigs',True)
        self.add_attribute('print_format','png')
        self.add_attribute('print_framenos','all')  # which frames to plot
        self.add_attribute('print_gaugenos','all')  # which gauges to plot
        self.add_attribute('print_fignos','all')    # which figures to plot each frame

        self.add_attribute('iplotclaw_fignos','all')    # which figures to plot interactively

        self.add_attribute('latex',True)                # make latex files for figures
        self.add_attribute('latex_fname','plots')       # name of latex file
        self.add_attribute('latex_title','Clawpack Results')
        self.add_attribute('latex_framesperpage','all') # number of frames on each page
        self.add_attribute('latex_framesperline',2)     # number of frames on each line
        self.add_attribute('latex_figsperline','all')   # number of figures on each line
        self.add_attribute('latex_makepdf',False)       # run pdflatex on latex file

        self.add_attribute('html',True)                # make html files for figures
        self.add_attribute('html_index_fname','_PlotIndex.html')   # name of html index file
        self.add_attribute('html_index_title','Plot Index')   # title at top of index page
        self.add_attribute('html_homelink',None)       # link to here from top of _PlotIndex.html
        self.add_attribute('html_movie','JSAnimation') # make html with java script for movie
        self.add_attribute('html_movie_width', 500)    # width of movie (not used?)
        self.add_attribute('html_movie_dpi', 100)      # dpi of movie

        self.add_attribute('html_eagle',False)         # use EagleClaw titles on html pages?

        self.add_attribute('kml',False)                # make kml plots and a kml file for figures
        self.add_attribute('kml_index_fname','_GoogleEarth')   # name of html index file
        self.add_attribute('kml_publish',None)
        self.add_attribute('kml_name',"GeoClaw")
        self.add_attribute('kml_starttime',None)
        self.add_attribute('kml_tz_offset',None)
        self.add_attribute('kml_time_scale',1.0) # Multiply by factor to get seconds
        self.add_attribute('kml_map_topo_to_latlong',None)
        self.add_attribute('kml_user_files',[])

        self.add_attribute('gif_movie',False)          # make animated gif movie of frames

        self.add_attribute('mp4_movie',False)         # make animated mp4 movie of frames
        self.add_attribute('movie_name_prefix', 'movie_')
        self.add_attribute('setplot',False)            # Execute setplot.py in plot routine

        self.add_attribute('mapc2p',None)              # function to map computational
                                        # points to physical


        self.add_attribute('beforeframe',None)         # function called before all plots
                                        # in each frame are done
        self.add_attribute('afterframe',None)          # function called after all plots
                                        # in each frame are done

        self.add_attribute('plotfigure_dict',{})
        try:
            from collections import OrderedDict   # new in Python 2.7
            d = OrderedDict()
        except:
            d = {}
        self.add_attribute('otherfigure_dict',d)

        self.add_attribute('framesoln_dict',{})        # dictionary for holding framesoln
                                        # objects associated with plots

        self.add_attribute('gaugesoln_dict',{})        # dictionary for holding gaugesoln
                                        # objects associated with plots

        self.add_attribute('save_frames',True)         # True ==> Keep a copy of any frame
                                        # read in.  False ==> Clear the frame
                                        # solution dictionary before adding
                                        # another solution

        self.add_attribute('save_figures',True)        # True ==> Keep a copy of and figure
                                        # created.  False ==> Clear the
                                        # figure dictionary before adding
                                        # another solution

        self.add_attribute('refresh_gauges',False)     # False ==> don't re-read gaugesoln if
                                        # already in gaugesoln_dict

        self.add_attribute('timeframes_framenos',None)
        self.add_attribute('timeframes_frametimes',None)
        self.add_attribute('timeframes_fignos',None)
        self.add_attribute('timeframes_fignames',None)



        self.add_attribute('gauges_gaugenos',None)
        self.add_attribute('gauges_fignos',None)
        self.add_attribute('gauges_fignames',None)

        # Parallel capabilities
        # Run multiple processess dividing up the frames that need to be plotted
        self.add_attribute('parallel', False)
        # Default to OMP_NUM_THREADS available if defined
        self.add_attribute('num_procs', None)
        self.add_attribute('proc_frames', None)
        self.add_attribute('_parallel_todo', None)


        self._next_FIG = 1000
        self._fignames = []
        self._fignos = []
        self._mode = 'unknown'
        self._figname_from_num = {}
        self._otherfignames = []


    def new_plotfigure(self, name=None, figno=None, type='each_frame'):
        """
        Create a new figure for Clawpack plots.
        If type='each_frame' it is a figure that will be plotted
        for each time frame.
        If type='multi_frame' it is a figure that will be plotted based on
        all the frames, such as x-t plots or time series. (Not yet implemented)
        """
        if (self._mode != 'iplotclaw') and (name in self._fignames):
            print('*** Warning, figure named %s has already been created' % name)
        if (self._mode != 'iplotclaw') and (figno in self._fignos):
            print('*** Warning, figure number %s has already been created' % figno)
        if figno is None:
            self._next_FIG += 1
            figno = self._next_FIG
        if name is None:
            name = "FIG%s" % figno
        if name in self._fignames:
            print("*** Error in new_plotfigure: Figure name already used... ",name)
            raise Exception("Figure name already used")
        elif figno in self._fignos:
            print("*** Error in new_plotfigure: Figure number already used... ",figno)
            raise Exception("Figure number already used")

        self._fignames.append(name)
        self._fignos.append(figno)
        plotfigure = ClawPlotFigure(name, figno, type, self)
        if not self.save_figures:
            self.plotfigure_dict.clear()
        self.plotfigure_dict[name] = plotfigure
        self._figname_from_num[figno] = name
        return plotfigure


    def getframe(self,frameno,outdir=None,refresh=False):
        """
        ClawPlotData.getframe:
        Return an object of class Solution containing the solution
        for frame number frameno.

        If refresh == True then this frame is read from the fort
        files, otherwise it is read from the fort files only if the
        the dictionary self.framesoln_dict has no key frameno.  If it does, the
        frame has previously been read and the dictionary value is returned.
        """

        from clawpack.pyclaw import solution

        framesoln_dict = self.framesoln_dict

        if 0:
            if outdir:
                key = (frameno, outdir)
            else:
                key = frameno
                outdir = self.outdir

        if outdir is None:
            outdir = self.outdir
        outdir = os.path.abspath(outdir)
        key = (frameno, outdir)

        if refresh or (key not in framesoln_dict):
            framesoln = solution.Solution(frameno,path=outdir,
                                          file_prefix=self.file_prefix,
                                          file_format=self.format)
            if not self.save_frames:
                framesoln_dict.clear()
            framesoln_dict[key] = framesoln
            if key != frameno:
                print('    Reading  Frame %s at t = %g  from outdir = %s' \
                    % (frameno,framesoln.t,outdir))
            else:
                print('    Reading  Frame %s at t = %g  ' \
                    % (frameno,framesoln.t))
        else:
            framesoln = self.framesoln_dict[key]

        return framesoln


    def clearfigures(self):
        """
        Clear all plot parameters specifying figures, axes, items.
        Does not clear the frames of solution data already read in.
        For that use clearframes.
        """

        self.plotfigure_dict.clear()
        self._fignames = []
        self._fignos = []
        self._next_FIG = 1000
        self._otherfignames = []


    def clearframes(self, framenos='all'):
        """
        Clear one or more frames from self.framesoln_dict.
        Need to add outdir option!
        """

        if isinstance(framenos, int):
            framenos = [framenos]  # turn into a list

        if framenos=='all':
            self.framesoln_dict.clear()
            print('Cleared all frames')
        else:
            for frameno in framenos:
                xxx = self.plotdata.framesoln_dict.pop(frameno,None)
                if xxx is None:
                   print('No frame data to clear for frame ',frameno)
                else:
                   print('Cleared data for frame ',frameno)


    def getgauge(self, gauge_id, outdir=None, verbose=True):
        r"""Read in the gauge labeled with `gaugeno` in path `outdir`

        :Note:
        The behavior of this function has changed to actually only read in the
        requested gauge id rather than all of the gauges.  The dictionary
        `gaugesoln_dict` remains the same.

        :Input:
         - *gauge_id* - (int) The gauge id of the gauge to be read in.
         - *outdir* - (path) Path to output directory containing gauge files.
           Defaults to this data object's `self.outdir`.
         - *verbose* - (bool) Verbose console output, default is `False`.

        :Output:
         - (clawpack.amrclaw.GaugeSolution) The read in gauge solution either
           from the `gaugeson_dict` or from file.  If something went wrong then
           the routine prints a warning and returns `None`.
        """

        # Construct path to file
        if outdir is None:
            outdir = self.outdir
        outdir = os.path.abspath(outdir)

        # Reread gauge data file
        key = (gauge_id, outdir)
        if self.refresh_gauges or (key not in self.gaugesoln_dict):

            try:

                # Read gauge solution:
                import clawpack.pyclaw.gauges as gauges

                self.gaugesoln_dict[key] = gauges.GaugeSolution(
                                           gauge_id=gauge_id, path=outdir)

                if verbose:
                    print("Read in gauge %s." % gauge_id)

            except Exception as e:
                import warnings
                warnings.warn(str(e))
                return None

        return self.gaugesoln_dict[key]


    def plotframe(self, frameno):
        from clawpack.visclaw import frametools
        frametools.plotframe(frameno, self)

    def printframes(self, verbose=True):
        #from clawpack.visclaw import frametools
        #frametools.printframes(self, verbose)
        print("*** printframes is deprecated.  Use plotpages.plotclaw_driver")
        print("*** for added capabilities.")
        raise DeprecationWarning("The method 'printframes' is deprecated.")

    def fignos(self):
        """
        Return a list of the figure numbers actually used.
        Useful in afterframe function for example to loop over all
        figures and do something.
        """
        return self._fignos

    def mode(self):
        """
        Return self._mode, which is set internally to
           'iplotclaw' if Iplotclaw is in use,
           'printframes' if printframes is being used
        Useful in afterframe function if you want to do different things
           for interactive or print modes.
        """
        return self._mode

    def iplotclaw(self):
        """
        Return True if interactive plotting with iplotclaw is being done.
        """
        return (self._mode == 'iplotclaw')


    def getfigure(self,figname):
        try:
            plotfigure = self.plotfigure_dict[figname]
        except:
            raise Exception('Error accessing plotfigure_dict[%s]' % figname)

        return plotfigure

    def getaxes(self,axesname,figname=None):
        found = True
        if not figname:
            found = False
            for fig in self._fignames:
                plotfigure = self.getfigure(fig)
                if axesname in plotfigure._axesnames:
                    if found == True: # already found!
                        print('*** Ambiguous... must specify figname')
                        print('    try getaxes(axesname, figname)')
                        return None
                    figname = fig
                    found = True
        if not found:
            print('*** No axes found with name = ',axesname)
            return None
        try:
            plotfigure = self.getfigure(figname)
            plotaxes = plotfigure.plotaxes_dict[axesname]
        except:
            print('*** Error accessing plotaxes[%s]' % axesname)
            print('*** figname = %s' % figname)
            return None
        return plotaxes

    def getitem(self,itemname,axesname=None,figname=None):
        found = True
        if not figname:
            # search over all figures looking for the item
            found = False
            for fign in self._fignames:
                plotfigure = self.getfigure(fign)
                if not axesname:
                    # search over all axes looking for the item
                    for axesn in plotfigure._axesnames:
                        plotaxes = self.getaxes(axesn,fign)
                        if itemname in plotaxes._itemnames:
                            if found == True: # already found!
                                print('*** Ambiguous... must specify figname and/or axesname')
                                print('    try getitem(itemname, axesname, figname)')
                                return None
                            axesname = axesn
                            figname = fign
                            found = True
                else:
                    # axesname was specified (but not figname)
                    plotaxes = self.getaxes(axesname,fign)
                    if itemname in plotaxes._itemnames:
                        if found == True: # already found!
                            print('*** Ambiguous... must specify figname and/or axesname')
                            print('    try getitem(itemname, axesname, figname)')
                            return None
                        figname = fign
                        found = True

        elif not axesname:
            # figname was specified but not axesname.
            # search over all axes looking for the item
            found = False
            plotfigure = self.getfigure(figname)
            for axesn in plotfigure._axesnames:
                plotaxes = self.getaxes(axesn,figname)
                if itemname in plotaxes._itemnames:
                    if found == True: # already found!
                        print('*** Ambiguous... must specify axesname')
                        print('    try getitem(itemname, axesname, figname)')
                        return None
                    axesname = axesn
                    found = True

        if not found:
            print('*** No item found with name = ',itemname)
            return None
        try:
            plotaxes = self.getaxes(axesname,figname)
            plotitem = plotaxes.plotitem_dict[itemname]
        except:
            print('*** Error accessing plotitem[%s]' % itemname)
            print('*** figname = ',figname)
            print('*** axesname = ',axesname)
            return None
        return plotitem


    def showitems(self):
        fignames = self._fignames
        print("\n\nCurrent plot figures, axes, and items:")
        print("---------------------------------------")
        for figname in fignames:
            plotfigure = self.getfigure(figname)
            s =  "  figname = %s, figno = %s" % (figname, plotfigure.figno)
            if not plotfigure._show:
                s = s + "  [Not showing]"
            print(s)
            axesnames = plotfigure._axesnames
            for axesname in axesnames:
                plotaxes = self.getaxes(axesname,figname)
                s =  "     axesname = %s, axescmd = %s" \
                       % (axesname, plotaxes.axescmd)
                if not plotaxes._show:
                    s = s + "  [Not showing]"
                print(s)
                for itemname in plotaxes._itemnames:
                    plotitem = self.getitem(itemname,axesname,figname)
                    plot_type = plotitem.plot_type
                    s =  "        itemname = %s,  plot_type = %s" \
                          % (itemname,plot_type)
                    if not plotitem._show:
                        s = s + "  [Not showing]"
                    print(s)
            print(" ")


    def getq(self,frameno):
        solution = self.getframe(frameno)
        patches = solution.patches
        if len(patches) > 1:
            print('*** Warning: more than 1 patch, q on patch[0] is returned')
        q = patches[0].q
        return q


    def new_otherfigure(self, name=None, fname=None):
        """
        Create a new figure for Clawpack plots.
        For figures not repeated each frame.
        """
        if (self._mode != 'iplotclaw') and (name in self._fignames):
            print('*** Warning, figure named %s has already been created' % name)

        if name is None:
            if fname is None:
                raise Exception("Need to provide name in new_otherfigure")
            else:
                name = fname
        if name in self._otherfignames:
            print("*** Error in new_otherfigure: Figure name already used... ",name)
            raise Exception("Figure name already used")

        self._otherfignames.append(name)
        otherfigure = ClawOtherFigure(name,self)
        self.otherfigure_dict[name] = otherfigure
        otherfigure.fname = fname
        return otherfigure

    def set_outdirs(self):
        """
        Make a list of all outdir's for all plotitem's in the order they
        are first used.
        """

        outdir_list = []
        for figname in self._fignames:
            plotfigure = self.plotfigure_dict[figname]
            if not plotfigure._show:
                continue  # skip to next figure
            for axesname in plotfigure._axesnames:
                plotaxes = plotfigure.plotaxes_dict[axesname]
                if not plotaxes._show:
                    continue  # skip to next axes
                for itemname in plotaxes._itemnames:
                    plotitem = plotaxes.plotitem_dict[itemname]
                    if not plotitem._show:
                        continue  # skip to next item
                    if plotitem.outdir is not None:
                        outdir = plotitem.outdir
                    else:
                        outdir = self.outdir
                    if outdir not in outdir_list:
                        outdir_list.append(outdir)

        self._outdirs = outdir_list
        return self

# ============================================================================
#  Subclass ClawPlotFigure containing data for plotting a figure
# ============================================================================
class ClawPlotFigure(clawdata.ClawData):
    """

    Data subclass containing plot data needed to plot a single figure.
    This may consist of several ClawPlotAxes objects.

    """

    # ========================================================================
    #  Initialization routine
    # ========================================================================
    def __init__(self, name, figno, fig_type, plotdata):
        """
        Initialize a ClawPlotFigure object
        """

        super(ClawPlotFigure, self).__init__()

        self._plotdata = plotdata           # parent ClawPlotData object
        self.add_attribute('name',name)
        self.add_attribute('figno',figno)
        self.add_attribute('figsize',None)
        self.add_attribute('facecolor',None)
        self.add_attribute('kwargs',{})
        self.add_attribute('clf_each_frame',True)
        self.add_attribute('clf_each_gauge',True)
        self._axesnames = []
        self.add_attribute('show',True)
        self._show = True
        self.add_attribute('plotaxes_dict', {})
        self.add_attribute('type',fig_type)   # = 'each_frame' or 'each_run' or 'each_gauge'
        self.add_attribute('use_for_kml',False)
        self.add_attribute('kml_gauge_name','Gauge')
        self.add_attribute('kml_dpi',200)
        self.add_attribute('kml_xlimits',None)
        self.add_attribute('kml_ylimits',None)
        self.add_attribute('kml_use_figure_limits',True)
        self.add_attribute('kml_tile_images',False)
        self.add_attribute('kml_colorbar',None)
        self.add_attribute('kml_use_for_initial_view',False)
        self.add_attribute('kml_show_figure',False)
        self.add_attribute('kml_maxlevel',20)
        self.add_attribute('kml_figsize',None)    # Figure size;  specify to get rid of aliasing
        self._next_AXES = 0

    def new_plotaxes(self, name=None, type='each_frame'):
        """
        Create a new axes that will be plotted in this figure.
        If type='each_frame' it is an axes that will be plotted
        for each time frame.
        If type='multi_frame' it is an axes that will be plotted based on
        all the frames, such as x-t plots or time series. (Not yet implemented)
        If type='empty' it is created without doing any plots using the
        pyclaw tools.  Presumably the user will create a plot within an
        afteraxes command, for example.
        """
        if name is None:
            self._next_AXES += 1
            name = "AXES%s" % self._next_AXES
        if name in self._axesnames:
            print('*** Warning, axes named %s has already been created' % name)

        if name not in self._axesnames:
            self._axesnames.append(name)
        plotaxes = ClawPlotAxes(name, self)
        self.plotaxes_dict[name] = plotaxes
        plotaxes.type = type
        return plotaxes

    def gethandle(self):
        _handle = getattr(self,'_handle',None)
        return _handle


# ============================================================================
#  Subclass ClawPlotAxes containing data for plotting axes within a figure
# ============================================================================
class ClawPlotAxes(clawdata.ClawData):
    """

    Data subclass containing plot data needed to plot a single axes.
    This may consist of several ClawPlotItem objects.

    """

    # ========================================================================
    #  Initialization routine
    # ========================================================================
    def __init__(self, name, plotfigure):
        """
        Initialize a ClawPlotAxes object
        """

        super(ClawPlotAxes, self).__init__()

        self._plotfigure = plotfigure                   # figure this item is on
        self._plotdata = plotfigure._plotdata           # parent ClawPlotData object

        self.add_attribute('name',name)
        self.add_attribute('title',name)
        self.add_attribute('title_with_t',True)  # creates title of form 'title at time t = ...'
        self.add_attribute('axescmd','subplot(1,1,1)')

        self.add_attribute('beforeaxes',None)
        self.add_attribute('afteraxes',None)
        self.add_attribute('xlimits',None)
        self.add_attribute('ylimits',None)
        self.add_attribute('skip_patches_outside_xylimits',None)
        self.add_attribute('scaled',False)   # true so x- and y-axis scaled same
        self.add_attribute('image',False)    # true so x- and y-axis scaled same
                                             # and plot bounds tight
        self.add_attribute('plotitem_dict', {})
        self.add_attribute('type','each_frame')
        self._itemnames = []
        self.add_attribute('show',True)
        self._show = True
        self._handle = None
        self._next_ITEM = 0
        self.add_attribute('figno', self._plotfigure.figno)

        # attributes for gauge plots
        self.add_attribute('time_label', 'time')  # for time axis in gauges
        self.add_attribute('time_label_fontsize', None)
        self.add_attribute('time_label_kwargs', {})  # kwargs for xlabel cmd
        self.add_attribute('time_scale', 1)  # multiplicative factor to rescale t
                                             # e.g. 1/3600. from sec to hours

        # recently added attributes:
        self.add_attribute('kwargs', {})
        self.add_attribute('grid', None) # True to add grid() command
        self.add_attribute('grid_kwargs', {}) # argument to grid() command
        self.add_attribute('title_fontsize', None)
        self.add_attribute('title_kwargs', {}) # e.g. to set color
        self.add_attribute('title_t_format', None) # format for t in title
        self.add_attribute('xticks_fontsize', None)
        self.add_attribute('xticks_kwargs', {}) # e.g. to set ticks,rotation
        self.add_attribute('yticks_fontsize', None)
        self.add_attribute('yticks_kwargs', {}) # e.g. to set ticks
        self.add_attribute('xlabel', None) # label for x-axis
        self.add_attribute('ylabel', None) # label for y-axis
        self.add_attribute('xlabel_fontsize', None)
        self.add_attribute('ylabel_fontsize', None)
        self.add_attribute('xlabel_kwargs', {})
        self.add_attribute('ylabel_kwargs', {})
        self.add_attribute('aspect', None)
        self.add_attribute('aspect_latitude', None)
        self.add_attribute('useOffset', None)


    def new_plotitem(self, name=None, plot_type=None):
        # Create a new entry in self.plotitem_dict

        if name is None:
            self._next_ITEM += 1
            name = "ITEM%s" % self._next_ITEM

        if name not in self._itemnames:
            self._itemnames.append(name)

        plotitem = ClawPlotItem(name, plot_type, plotaxes=self)

        self.plotitem_dict[name] = plotitem

        return plotitem

    def get_plotdata(self):
        plotdata = getattr(self,'_plotdata',None)
        return self._plotdata

    def get_plotfigure(self):
        plotfigure = getattr(self,'_plotfigure',None)
        return self._plotfigure

    def gethandle(self):
        _handle = getattr(self,'_handle',None)
        return self._handle

# ============================================================================
#  Subclass ClawPlotItem containing data for plotting a single object
# ============================================================================
class ClawPlotItem(clawdata.ClawData):
    """

    Data subclass containing plot data needed to plot a single object.
    This may be a single curve, set of points, contour plot, etc.

    """

    # ========================================================================
    #  Initialization routine
    # ========================================================================
    def __init__(self, name, plot_type, plotaxes):
        """
        Initialize a ClawPlotItem object
        """

        super(ClawPlotItem, self).__init__()


        self._plotaxes = plotaxes                       # axes this item is on
        self._plotfigure = plotaxes._plotfigure         # figure this item is on
        self._plotdata = plotaxes._plotfigure._plotdata           # parent ClawPlotData object

        try:
            num_dim = int(plot_type[0])   # first character of plot_type should be num_dim
        except:
            print('*** Error: could not determine num_dim from plot_type = ',plot_type)

        self.add_attribute('num_dim',num_dim)
        self.add_attribute('name',name)
        self.add_attribute('figno',plotaxes.figno)

        self.add_attribute('outdir',None)              # indicates data comes from
                                        #   self._plotdata.outdir

        self.add_attribute('plot_type',plot_type)
        self.add_attribute('plot_var',0)
        self.add_attribute('data_show',True)

        self.add_attribute('MappedGrid',None)          # False to plot on comput. patch even
                                        # if _plotdata.mapc2p is not None.

        self.add_attribute('mapc2p',None)              # function to map computational
                                    # points to physical (over-rides
                                    # plotdata.mapc2p if set for item


        self.add_attribute('afterpatch',None)           # function called after each patch is
                                        # plotted within each single plotitem.
        self.add_attribute('afteritem',None)           # function called after the item is
                                        # plotted for each frame

        self.add_attribute("show",True)                # False => suppress showing this item
        self._show = True               # Internal

        self._current_pobj = None

        self.add_attribute('params',{})  # dictionary to hold optional parameters

        if num_dim == 1:
            self.add_attribute('plotstyle','-')
            self.add_attribute('color',None)
            self.add_attribute('kwargs',{})
            amr_attributes = """show color kwargs data_show plotstyle""".split()
            for a in amr_attributes:
                self.add_attribute('amr_%s' % a, [])

            if plot_type == '1d_fill_between':
                zero_function = lambda current_data: 0.
                self.add_attribute('plot_var2',zero_function)
                self.add_attribute('fill_where',None)

            if plot_type == '1d_from_2d_data':
                self.add_attribute('map_2d_to_1d',None)
                self.add_attribute('amr_plotstyle',[])

        elif num_dim == 2:

            # default values specifying this single plot:
            self.add_attribute('plot_type',plot_type)
            self.add_attribute('celledges_show',0)
            self.add_attribute('celledges_color','k')
            self.add_attribute('celledges_linewidth',0.5)
            self.add_attribute('patch_bgcolor','w')
            self.add_attribute('patchedges_show',0)
            self.add_attribute('patchedges_color','k')
            self.add_attribute('patchedges_linewidth',0.8)
            self.add_attribute('add_colorbar',False)
            self.add_attribute('colorbar_shrink',None)
            self.add_attribute('colorbar_label',None)
            self.add_attribute('colorbar_ticks', None)
            self.add_attribute('colorbar_tick_labels',None)
            self.add_attribute('colorbar_extend',None)
            self.add_attribute('colorbar_kwargs',{})
            self.add_attribute('kwargs',{})
            amr_attributes = """celledges_show celledges_color 
              celledges_linewidth data_show patch_bgcolor patchedges_show 
              patchedges_color kwargs""".split()
            for a in amr_attributes:
                self.add_attribute('amr_%s' % a, [])

            if plot_type == '2d_pcolor':
                from clawpack.visclaw import colormaps
                self.add_attribute('pcolor_cmap',colormaps.yellow_red_blue)
                self.add_attribute('pcolor_cmin',None)
                self.add_attribute('pcolor_cmax',None)

            elif plot_type == '2d_imshow':
                from clawpack.visclaw import colormaps
                self.add_attribute('imshow_cmap',colormaps.yellow_red_blue)
                self.add_attribute('imshow_cmin',None)
                self.add_attribute('imshow_cmax',None)
                self.add_attribute('imshow_norm', None)
                self.add_attribute('imshow_alpha', None)


            elif plot_type in ['2d_contour', '2d_contourf']:
                self.add_attribute('contour_nlevels',20)
                self.add_attribute('contour_levels',None)
                self.add_attribute('contour_min',None)
                self.add_attribute('contour_max',None)
                self.add_attribute('contour_show',1)
                self.add_attribute('contour_colors','k')
                self.add_attribute('contour_cmap',None)
                amr_attributes = """show colors cmap data_show""".split()
                for a in amr_attributes:
                    self.add_attribute('amr_contour_%s' % a, [])


                if plot_type == '2d_contourf':
                    self.add_attribute('fill_cmap',None)
                    self.add_attribute('fill_cmin',None)
                    self.add_attribute('fill_cmax',None)
                    self.add_attribute('fill_colors',None)
                    # Note either fill_cmap or fill_colors must be None

            elif plot_type == '2d_schlieren':
                from clawpack.visclaw import colormaps
                self.add_attribute('schlieren_cmap',colormaps.schlieren_grays)
                self.add_attribute('schlieren_cmin',None)
                self.add_attribute('schlieren_cmax',None)

            elif plot_type == '2d_patch':
                self.add_attribute('max_density',None)
                self.celledges_show = True
                self.patchedges_show = True

            elif plot_type == '2d_quiver':
                self.add_attribute('quiver_var_x',None)
                self.add_attribute('quiver_var_y',None)
                self.add_attribute('quiver_coarsening',1)
                self.add_attribute('quiver_key_show',False)
                self.add_attribute('quiver_key_label_x',0.15)
                self.add_attribute('quiver_key_label_y',0.95)
                self.add_attribute('quiver_key_units','')
                self.add_attribute('quiver_key_scale',None)
                self.add_attribute('quiver_key_kwargs',{})
                amr_attributes = """coarsening key_show key_label_x key_label_y
                         key_scale key_kwargs data_show""".split()
                for a in amr_attributes:
                    self.add_attribute('amr_quiver_%s' % a, [])

            elif plot_type == '2d_hillshade':
                self.add_attribute('hillshade_cmap','gray')
                self.add_attribute('hillshade_vertical_exaggeration',1)
                self.add_attribute('hillshade_azimuth_degree',315)
                self.add_attribute('hillshade_altitude_degree',45)
                self.add_attribute('hillshade_latlon', False)

            else:
                 print('*** Warning 2d plot type %s not recognized' % plot_type)

        elif num_dim == 3:
            raise NotImplementedError('ClawPlotItem not yet set up for num_dim = 3')

        else:
            raise Warning('Unrecognized plot_type in ClawPlotItem')


    def getframe(self,frameno,refresh=False):
        """
        ClawPlotItem.getframe:
        Return an object of class Solution containing the solution
        for frame number frameno.

        If refresh == True then this frame is read from the fort
        files, otherwise it is read from the fort files only if the
        the dictionary self.framesoln_dict has key frameno.  If it does, the
        frame has previously been read and the dictionary value is returned.
        """

        plotdata = self._plotdata
        outdir = self.outdir
        framesoln = plotdata.getframe(frameno, outdir,refresh=refresh)

        return framesoln


    def getgauge(self,gauge):
        """
        ClawPlotItem.getgauge:
        Return an object of class GaugeSolution containing the solution
        for gauge number gaugeno.

        If self.refresh_gauges == True then this gauge is read from the
        fort.gauge file, otherwise it is read only if the
        the dictionary self.gaugesoln_dict has no key gaugeno.  If it does, the
        gauge has previously been read and the dictionary value is returned.
        """

        plotdata = self._plotdata
        outdir = self.outdir
        gaugesoln = plotdata.getgauge(gauge, outdir)

        return gaugesoln


# ============================================================================
#  Subclass ClawOtherFigure containing data for plotting a figure
# ============================================================================
class ClawOtherFigure(clawdata.ClawData):
    """

    Data subclass containing plot data needed to plot a single figure.
    For figures that are not produced each frame.

    """

    # ========================================================================
    #  Initialization routine
    # ========================================================================
    def __init__(self, name, plotdata):
        """
        Initialize a ClawOtherFigure object
        """

        super(ClawOtherFigure, self).__init__()

        self._plotdata = plotdata           # parent ClawPlotData object
        self.add_attribute('name',name)
        self.add_attribute('fname',None)            # name of png file
        self.add_attribute('makefig',None)          # function invoked to create figure



#!/usr/bin/python

# Author: Xinsheng (Shawn) Qin
# date: 03/2016
# Modified by Katy Barnhart
# date 05/2020

import sys
import os
import numpy as np
from clawpack.pyclaw import Solution

import logging

logger = logging.getLogger('pyclaw.fileio')

try:
    import vtk
except ImportError:
    logging.critical("Could not import vtk!")
    error_msg = ("Could not import VTK, please install (package are available"
    " through conda-forge and pypi. See the docstring for details).")
    print(error_msg)

from vtk.util import numpy_support

from vtk import (
    vtkOverlappingAMR,
    vtkUniformGrid,
    vtkAMRBox,
    vtkXMLUniformGridAMRWriter)

def write(
    solution,
    frame,
    path="_output",
    file_prefix='claw',
    write_aux=None,
    options=None,
    write_p=None,
    ):
    """Write out a VTK representation of solution

    This capability requires installation of the vtk module that is created by
    Kitware. Binary distributions are available through
    `PyPI <https://pypi.org/project/vtk/>`_
    or `conda-forge <https://anaconda.org/conda-forge/vtk>`_.

    .. code-block:: bash

       $ conda install -c conda-forge vtk

    or

    .. code-block:: bash

       $ pip install vtk

    For each input frame the following files and directories are created in
    the directory specified by *path*:

      - input_prefixXXXX.vthb. This file provides the metadata to
        describe how AMR patches are represented in the .vti files
        (including a relative
        path to the .vti files).
      - directory: input_prefixXXXX containing files called
        input_prefixXXXX_<index>.vti. <index> represents the file
        index. There is a file for each patch at each AMR level.

    Writing this function took advantage of the following VTK resources:
      - `the VTK users guide <https://www.kitware.com/products/books/VTKUsersGuide.pdf>`_
      - `this blog post on numpy integration <https://blog.kitware.com/improved-vtk-numpy-integration-part-5/>`_
      - `the vtkOverlappingAMR example <https://lorensen.github.io/VTKExamples/site/Python/CompositeData/OverlappingAMR/>`_
      - `the XML writing example <https://lorensen.github.io/VTKExamples/site/Python/IO/WriteXMLLinearCells/>`_

    This function uses the following VTK classes:

      - `vtkUniformGrid <https://vtk.org/doc/nightly/html/classvtkUniformGrid.html>`_
      - `vtkOverlappingAMR <https://vtk.org/doc/nightly/html/classvtkOverlappingAMR.html>`_
      - `vtkAMRBox <https://vtk.org/doc/nightly/html/classvtkAMRBox.html>`_
      - `vtkXMLUniformGridAMRWriter <https://vtk.org/doc/nightly/html/classvtkXMLUniformGridAMRWriter.html>`_

    To open in paraview, choose the group of .vthb files, not the group of
    folders. This will be read in as cell data. In order to use filters like
    WarpByScalar you must use the CellDataToPointData filter first.

    :Input:
     - *solution* - (:class:`~pyclaw.solution.Solution`) Pyclaw object to be
       output
     - *frame* - (int) Frame number
     - *path* - (string) Root path
     - *file_prefix* - (string) Prefix for the file name. ``default = 'claw'``
     - *write_aux* - (bool) Not implemented.
     - *options* - (dict) if contains the key value pair ``binary = True`` then
       output will be in binary rather than ascii. Default is ascii. Not implemented.
     - *write_p* - (bool) Not implemented.

    Note that some keyword arguments are not used. This is to maintain
    compatibility with the function signature expected by
    :py:class:`~pyclaw.Solution`

    Notes on what is not yet implemented
        - Add options for writing aux files.
        - Consider making an equivalent vtk.read function.
    """
    # get options from the options dictionary.
    binary = options.get("binary", False)

    # check types.
    assert(isinstance(frame, int))
    assert(isinstance(solution, Solution))

    # calculate overlapped status, used to identify some cells as ghosts.
    _set_overlapped_status(solution)

    global_origin = solution.state.patch.lower_global + [0.]  # base patch
    levels = [state.patch.level-1 for state in solution.states]

    # shift base level to 0, since the base level in clawpack
    # is 1 while the base level in VTK is 0
    level_count = {}
    level_spacing = {}  # spacing of each level
    for i, level in enumerate(levels):
        if level in level_count.keys():
            level_count[level] = level_count[level] + 1
        else:
            level_count[level] = 1
            spacing = solution.states[i].patch.delta
            spacing.append(spacing[0])  # dz = dx
            spacing = np.array(spacing)
            level_spacing[level] = spacing
    numLevels = len(level_count.keys())

    # a list of num of patches at each level
    blocksPerLevel = [item[1] for item in
                     sorted(level_count.items(),
                            key=lambda a: a[0])]

    # Initialize the vtkOverlappingAMR object. Provide it the number of levels,
    # number of blocks per level, and the global origin.
    amr = vtkOverlappingAMR()
    amr.Initialize(numLevels, blocksPerLevel)
    amr.SetOrigin(global_origin)

    # get states and initialize the global index (used below)
    states_sorted = sorted(solution.states, key=lambda a: a.patch.level)
    global_index = 0

    # for each AMR level create the vtkAMRBox and vtkUniformGrid and add to
    # the vtkOverlappingAMR object.
    for level in level_count.keys():
        # get number of blocks per level.
        nblocks = blocksPerLevel[level]

        # and the spacing at that level
        spacing = level_spacing[level]
        amr.SetSpacing(level, spacing)

        # for each block at this AMR level.
        for index in range(nblocks):
            # get the index used on the states_sorted file.
            local_index = global_index + index

            # get the origin and number of dimensions.
            origin = states_sorted[local_index].patch.lower_global + [0.]
            node_dims = [x + 1 for x in states_sorted[local_index].patch.num_cells_global + [0]]

            # create a vtkUniformGrid using the vtkAMRBox
            grid = vtkUniformGrid()
            grid.Initialize()
            grid.SetOrigin(origin)
            grid.SetSpacing(spacing)
            grid.SetDimensions(node_dims)

            # Construct an vtkAMRbox
            # Note that the dimensions specify the node dimensions, rather than the cell dimensions.
            # Nodes are one more than the cells.
            box = vtkAMRBox(origin, node_dims, spacing, global_origin)

            # Set the data of the vtkUniformGrid

            # get the cell data, and add each to the uniform grid.
            # ignore the last element of q, which provides info about overlapping.
            # it is set next.
            q = states_sorted[local_index].q

            for i in range(q.shape[0]-1):
                array_name = "q_"+str(i)
                q_i = q[i, ...]
                q_i = q_i.transpose()

                #https://pyscience.wordpress.com/2014/09/06/numpy-to-vtk-converting-your-numpy-arrays-to-vtk-arrays-and-files/
                # transform into an array.
                array = numpy_support.numpy_to_vtk(num_array=q_i.ravel(), deep=True, array_type=vtk.VTK_FLOAT)

                # set the name.
                array.SetName(array_name)

                # verify the sizes are correct.
                assert q_i.size == grid.GetNumberOfCells()

                # add the array to the uniform grid.
                grid.GetCellData().AddArray(array)

            # mark overlapping cells using the vtkGhostType array name.
            q_ol = q[-1, ...]  # last piece is used to mark overlapped cells
            q_ol = q_ol.transpose()
            array = numpy_support.numpy_to_vtk(q_ol.ravel(), deep=True, array_type=vtk.VTK_UNSIGNED_CHAR)
            array.SetName("vtkGhostType")
            # add the array to the uniform grid.
            grid.GetCellData().AddArray(array)

            # add AMR box and uniform grid to the overlapping AMR object.
            amr.SetAMRBox(level, index, box)
            amr.SetDataSet(level, index, grid)

            # verify that the box is not invalid.
            assert not box.IsInvalid()

        # after each level is done increment global_index used with states_sorted
        global_index += nblocks

    # write out the vtkOverlappingAMR object.
    out = os.path.join(path, file_prefix+str(frame).zfill(4)+'.vthb')
    writer = vtkXMLUniformGridAMRWriter()
    if not binary:
        writer.SetDataModeToAscii()
    writer.SetFileName(out)
    writer.SetInputData(amr)
    success = writer.Write()

    # assert writing returned 1, indicating success.
    assert success == 1


def _set_overlapped_status(sol):
    """
    return a list, overlapped_states,
    whose entries denote overlapped status for each patch.

    @type sol:  pyclaw.Solution
    @param sol: Solution object of pyclaw that contains all information
                of this time step.
    @rtype:     list
    @return:    add a component to the solution q,
                which contains overlapped status of each patch

    @type level_count:       dictionary
    @variable level_count:   a dictionary that maps levels
                             to number of patches of certain levels
                             e.g. {0:1, 1:2, 2:12}
    @type num_levels:        int
    @variable num_levels:    number of levels in total
    @type box_per_level:     list
    @variable box_per_level: [number of patches on level 0,
                              number of patches on level1, ...]

    """
    levels = [state.patch.level-1 for state in sol.states]
    # shift base level to 0
    level_count = {}
    level_spacing = {}  # spacing of each level
    for i, level in enumerate(levels):
        if level in level_count.keys():
            level_count[level] = level_count[level] + 1
        else:
            level_count[level] = 1
            spacing = sol.states[i].patch.delta
            spacing.append(spacing[0])  # dz = dx
            spacing = np.array(spacing)
            level_spacing[level] = spacing

    # a list of num of patches at each level
    box_per_level = [item[1] for item in
                     sorted(level_count.items(),
                            key=lambda a: a[0])]
    box_per_level = np.array(box_per_level)

    for state in sol.states:
        level = state.patch.level-1
        xlower_coarse = state.patch.dimensions[0].lower
        # xupper_coarse = state.patch.dimensions[0].upper
        ylower_coarse = state.patch.dimensions[1].lower
        # yupper_coarse = state.patch.dimensions[1].upper
        dx = state.patch.delta[0]
        dy = state.patch.delta[1]
        nx = state.patch.num_cells_global[0]
        ny = state.patch.num_cells_global[1]
        # in overlapped_status, entry with value 0 denotes
        # that the cell is not overlapped
        # entry with value 8 denotes that the cell is overlapped
        overlapped_status = np.zeros((1, state.q.shape[1], state.q.shape[2]))
        # convert from Fortran-Style to C-style
        # overlapped_status = overlapped_status.transpose(0, 2, 1)
        # In the future, efficiency of this part can be improved
        # by mapping grid levels to
        # a list of states of corresponding levels.
        # Otherwise, we need to scan each states in each outer loop as below
        for state_fine in sol.states:
            # find states with grid level of one higher
            if ((state_fine.patch.level-1) == level + 1):
                xlower_fine = state_fine.patch.dimensions[0].lower
                xupper_fine = state_fine.patch.dimensions[0].upper
                ylower_fine = state_fine.patch.dimensions[1].lower
                yupper_fine = state_fine.patch.dimensions[1].upper
                x_idx_lower = \
                    max(int(round((xlower_fine - xlower_coarse) /
                            float(dx))), 0)
                x_idx_upper = \
                    min(int(round((xupper_fine-xlower_fine)/dx)) +
                        x_idx_lower, nx)
                y_idx_lower = \
                    max(int(round((ylower_fine - ylower_coarse) /
                            float(dy))), 0)
                y_idx_upper = \
                    min(int(round((yupper_fine-ylower_fine)/dy)) +
                        y_idx_lower, ny)
                # set these cells to 8
                overlapped_status[0, x_idx_lower:x_idx_upper,
                                  y_idx_lower:y_idx_upper].fill(8)

            else:
                continue
        # state.q is in fortran style
        # state.q = np.vstack((state.q, overlapped_status.transpose(0, 2, 1)))
        state.q = np.vstack((state.q, overlapped_status))



#!/usr/bin/env python

from __future__ import print_function
import os
import numpy
import nose
import os
import warnings
from clawpack.geoclaw import topotools


# Set local test directory to get local files
testdir = os.path.dirname(__file__)
if len(testdir) == 0:
     testdir = "./"

extent = [-125,-124, 48, 48.5]

def test_etopo1_topo(make_plot=False, save=False):
    
    try:
        import netCDF4
    except ImportError:
        raise nose.SkipTest("netCDF4 not installed, skipping test")

    try:
        topo1 = topotools.read_netcdf('etopo1', extent=extent, verbose=True)
        topo10 = topotools.read_netcdf('etopo1', extent=extent, 
                                       coarsen=10, verbose=True)
    except (OSError, RuntimeError):
        warnings.warn('Could not read etopo1 data, check if thredds server up')
        raise nose.SkipTest("Reading etopo1 failed, skipping test")

    testdata_path = os.path.join(os.path.dirname(__file__), 'data', 'etopo1_10min.asc')
    if save:
        topo10.write(testdata_path, topo_type=3, Z_format='%.0f')
        print('Created %s' % testdata_path)

    topo10input = topotools.Topography()
    topo10input.read(testdata_path, topo_type=3)
    
    assert numpy.allclose(topo10.Z, topo10input.Z), \
           "topo10.Z does not agree with archived data"
    
    if make_plot:
        import matplotlib.pyplot as plt
        plt.figure(figsize=(12,5))
        ax1 = plt.subplot(1,2,1)
        topo1.plot(axes=ax1)
        plt.title('1 minute etopo1 data')
        ax10 = plt.subplot(1,2,2)
        topo10.plot(axes=ax10)
        plt.title('10 minute etopo1 data')
        pname = 'etopo1_test_plot.png'
        plt.savefig(pname)
        print('Created %s' % pname)
    
def test_etopo1_xarray():

    try:
        import xarray
    except ImportError:
        raise nose.SkipTest("xarray not installed, skipping test")
        
    try:
        topo10,topo10_xarray = topotools.read_netcdf('etopo1', extent=extent, 
                                                     return_xarray=True,
                                                     coarsen=10, verbose=True)
    except (OSError, RuntimeError):
        warnings.warn('Could not read etopo1 data, check if thredds server up')
        raise nose.SkipTest("Reading etopo1 failed, skipping test")

    testdata_path = os.path.join(testdir, 'data', 'etopo1_10min.asc')
    topo10input = topotools.Topography()
    topo10input.read(testdata_path, topo_type=3)
    
    assert numpy.allclose(topo10_xarray['z'], topo10input.Z), \
           "topo10_xarray['z'] does not agree with archived data"
    

if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        if "plot" in sys.argv[1].lower():
            test_etopo1_topo(make_plot=True)
        elif bool(sys.argv[1]):
            test_etopo1_topo(save=True)
    else:
        # Run tests
        test_etopo1_topo()
        test_etopo1_xarray()
        print("All tests passed.")




"""
Compare plots in all subdirectories to archived results.

Assumes the code has been freshly run in all subdirectories.  The script
make_all.py can be used for this.

Function test_subdirs returns True if all .png files agree in 
alls subdirectories the two directories test_dir and compare_dir.

Image differences can be viewed by opening a browser to 
    _image_diff/_ImageDiffIndex.html

If compare_dir is not specified, attempt to compare with plots in the latest
Clawpack gallery.  For this to work you need to first clone and/or
fetch a zip file of the latest gallery results from
     git://github.com/clawpack/clawpack.github.com
Put this in the top $CLAW directory so that there is a subdirectory
$CLAW/clawpack.github.com.  The archived plots can then be found in
$CLAW/clawpack.github.com/doc/_static.

"""

import os, subprocess
from clawpack.clawutil import imagediff


def test_subdirs(compare_dir=None,examples_dir='.',\
                 verbose=True,relocatable=False):

    from .make_all import list_examples

    if compare_dir is None:
        print("Attempting to find gallery for comparison...")
        try:
            CLAW = os.environ['CLAW']
        except:
            raise Exception("Need to set CLAW environment variable")
        examples_dir = os.path.abspath(examples_dir)

        iclaw = examples_dir.find(CLAW)
        if iclaw==-1:
            print("*** examples_dir is not under $CLAW, failed to find gallery")
            return False
              
        gallery_home = os.path.join(CLAW,'clawpack.github.com','doc','_static')
        package = os.path.split(os.path.split(examples_dir)[0])[1]
        compare_dir = os.path.join(gallery_home,package,'examples')
        if not os.path.isdir(compare_dir):
            print("*** Failed to find ",compare_dir)
            return False


    all_ok = True

    dir_list = list_examples(examples_dir)
    print("Found the following example subdirectories:")
    for d in dir_list:
        print("    ", d)
 
    print("Will run imagediff on _plots in the above subdirectories of ")
    print("    ", examples_dir)
    print("Will compare to _plots in corresponding subdirectories of: \n   %s" \
            % compare_dir)
    if 0:
        ans = input("Ok? ")
        if ans.lower() not in ['y','yes']:
            print("Aborting.")
            all_ok = False
            return all_ok

    top_dir = os.getcwd()
    for test_subdir in dir_list:
        print("\n=============================================================")
        print(test_subdir)
        print("=============================================================")
        #subdir = os.path.split(test_subdir)[1]
        subdir = test_subdir[len(examples_dir)+1:]
        compare_subdir = os.path.join(compare_dir, subdir)
        test_plots = os.path.join(test_subdir,'_plots')
        compare_plots = os.path.join(compare_subdir, '_plots')

        if not os.path.isdir(test_plots):
            print("*** Cannot find _plots directory ")
            print("*** Looking for ",test_plots)
            all_ok = False
            continue

        if not os.path.isdir(compare_plots):
            print("*** Cannot find _plots directory to compare against")
            print("*** Looking for ",compare_plots)
            all_ok = False
            continue
    
        os.chdir(subdir)
        try:
            regression_ok = imagediff.imagediff_dir(test_plots,compare_plots, \
                                    relocatable=relocatable,overwrite=True, \
                                    verbose=verbose)
            if not regression_ok:
                print("*** Regression files are not all identical in ")
                print("   ",test_subdir)
        except:
            print("*** Error running imagediff with directories \n  %s\n  %s\n" \
                        % (test_plots,compare_plots))
            regression_ok = False
        all_ok = all_ok and regression_ok
        os.chdir(top_dir)

    return all_ok
    
if __name__=="__main__":
    import sys
    all_ok = test_subdirs(*sys.argv[1:])
    print("==> all_ok = ",all_ok)





from __future__ import absolute_import
def setplot(plotdata):
    plotdata.clearfigures()

    # Figures corresponding to Figure 9.5 of LeVeque, "Finite Volume
    # Methods for Hyperbolic Problems," 2002 (though more of them)

    # Tuples of (variable name, variable number)
    figdata = [('Pressure', 0),
               ('Velocity', 1)]

    # Afteraxes function: draw a vertical dashed line at the interface
    # between different media
    def draw_interface(current_data):
        import pylab
        pylab.plot([0., 0.], [-1000., 1000.], 'k--')

    for varname, varid in figdata:
        plotfigure = plotdata.new_plotfigure(name=varname, figno=varid)

        plotaxes = plotfigure.new_plotaxes()
        plotaxes.xlimits = [-5., 5.]
        plotaxes.ylimits = [-0.5, 1.5]    # Good for both vars because of near-unit impedance
        plotaxes.title = varname
        plotaxes.afteraxes = draw_interface

        plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
        plotitem.plot_var = varid
        plotitem.color = 'b'

    plotdata.printfigs = True          # Whether to output figures
    plotdata.print_format = 'png'      # What type of output format
    plotdata.print_framenos = 'all'    # Which frames to output
    plotdata.print_fignos = 'all'      # Which figures to print
    plotdata.html = True               # Whether to create HTML files
    plotdata.latex = False             # Whether to make LaTeX output

    return plotdata




"""
Create a dtopo file for 1d geoclaw.
In this example a single fault is split into 3 subfaults for illustration.
If fault.rupture_type == 'kinematic' is used then all subfaults have the
same rupture time and rise time in this example.

The final seafloor deformation with fault.rupture_type == 'kinematic'
should be the same as if fault.rupture_type == 'static' is set, in which
case all subfaults rupture simultaneously and instantaneously at 1 second.
"""

from pylab import *

from clawpack.geoclaw.data import LAT2METER
from importlib import reload

from clawpack.geoclaw import dtopotools
reload(dtopotools)

fault = dtopotools.Fault1d()
fault.subfaults = []

strike = 0.  # 0 ==> top is at right in 1d,  180 ==> top at left
             # assuming 0 < dip <= 90

fault_top_meters = -100e3  # location x0 (in meters) at top edge in 1d
longitude0 = fault_top_meters/LAT2METER  # convert to degrees at latitude 0

fault_top_depth = 20e3     # depth of top below surface (meters)
width = 50e3  # width of fault in 1d (down-dip direction)

dip = 10.  # dip in degrees
theta = dip*pi/180.   # dip in radians


if 0:
    fault.rupture_type = 'static'  # instantaneous and simultaneous on subfaults
    rupture_time = 0.0
    rise_time = 0.
else:
    fault.rupture_type = 'kinematic'
    # use these same values for all subfaults in this simple example:
    rupture_time = 0.0
    rise_time = 20.  
      
average_slip = 10.0

# split into subfaults if desired:
nsubfaults = 3
max_slip = 2*average_slip # if modulated by cosine hump below
dlongitude = width*cos(theta)/LAT2METER / nsubfaults

ddepth = width*sin(theta) / nsubfaults
subfault_width = width/nsubfaults

total_slip = 0.  # keep track
for i in range(nsubfaults):
    # split total slip between subfaults, starting at top
    subfault = dtopotools.SubFault1d()
    subfault.dip = dip
    subfault.width = subfault_width
    subfault.depth = fault_top_depth + ddepth*i
    #subfault.slip = max_slip * 0.5*(1 - cos(2*pi*(i+0.5)/nsubfaults))
    subfault.slip = average_slip  # for constant slip
    total_slip += subfault.slip
    print('subfault %2i at depth %8.3f km has slip = %6.3f' \
            % (i,subfault.depth/1e3,subfault.slip))

    subfault.longitude = longitude0 + i*dlongitude
    subfault.coordinate_specification = 'top center'
    subfault.strike = strike
    subfault.rupture_time = rupture_time # all at same time
    subfault.rise_time = rise_time
    fault.subfaults.append(subfault)

print('average slip = %6.3f' % (total_slip/nsubfaults))


if fault.rupture_type == 'static':
    tend = 1.
    times = [0.,tend]
elif fault.rupture_type == 'kinematic':
    tend = 0.
    for s in fault.subfaults:
        tend = max(tend, s.rupture_time + 2*s.rise_time)
    times = linspace(0,tend,11)
    

print('dtopofile will have times: ',times)
xgrid,zgrid = loadtxt('celledges.data', skiprows=1, unpack=True)

# coarsen if desired:
xgrid = linspace(xgrid[0],xgrid[-1],100)

x = xgrid / LAT2METER  # convert meters to longitude
y = array([0,1])  # for 1d Okada

dtopo2d = fault.create_dtopography(x,y,times)
dtopo = dtopotools.DTopography1d()
dtopo.x = dtopo2d.x * LAT2METER # convert x back from degrees to meters
dtopo.times = dtopo2d.times
dtopo.dZ = dtopo2d.dZ[:,0,:]  # should be constant in y, remove that index

fname = 'dtopo_okada.dtt1'
dtopo.write(fname, dtopo_type=1)
print('Created ',fname)


if 1:
    # make plots
    figure(351,figsize=(6,8))
    clf()
    #dz = dtopo.dZ[-1,:]  # slice in x at final time
    #plot(dtopo.x,dz)
    subplot(211)
    for k in range(1,len(times)):
        plot(dtopo.x/1e3, dtopo.dZ[k,:],label='t = %6.2fs' % times[k])
    title('Okada seafloor deformation')
    legend(loc='upper right',fontsize=6)
    xticks(rotation=20)
    xlim(dtopo.x[0]/1e3,dtopo.x[-1]/1e3)
    xlabel('kilometers')
    ylabel('meters')
    grid(linewidth=0.5)
    
    subplot(212)
    for k,subfault in enumerate(fault.subfaults):
        xtop = subfault.longitude * LAT2METER
        xbottom = xtop + subfault.width*cos(subfault.dip*pi/180)
        ytop = -subfault.depth
        ybottom = ytop - subfault.width*sin(subfault.dip*pi/180)
        plot([xtop/1e3,xbottom/1e3],[ytop/1e3,ybottom/1e3], 
             label='subfault %i' % k)
    title('Subfaults')
    legend(loc='upper right')
    xlim(dtopo.x[0]/1e3,dtopo.x[-1]/1e3)
    xticks(rotation=20)
    xlabel('kilometers')
    ylabel('kilometers below seafloor')
    ylim(1.1*ybottom/1e3,0)
    grid(linewidth=0.5)
    tight_layout()

    fname = 'dtopo_okada.png'
    savefig(fname)
    print('Created ',fname)
    



num_waves = 3
num_eqn   = 4
num_aux   = 16

# Conserved quantities
depth = 0
x_momentum = 1
y_momentum = 2
z_momentum = 3

# Auxiliary variables
kappa = 0
enx = 1
eny = 2
enz = 3
etx = 4
ety = 5
etz = 6
enx = 7
eny = 8
enz = 9
etx = 10
ety = 11
etz = 12
erx = 13
ery = 14
erz = 15



import numpy as np
from pytest import importorskip
import os
import subprocess

thisdir = os.path.dirname(__file__)

class TestEulerGravity3D:

    def test_rising_hot_sphere(self):
        _ = importorskip("scipy") # skip test if scipy is not installed
        
        # Read expected solution
        expected_density = np.loadtxt(os.path.join(thisdir,'verify_rising_hot_sphere_classic_1.txt'))

        #Run test
        from . import rising_hot_sphere
        claw = rising_hot_sphere.euler3d(kernel_language='Fortran',solver_type='classic', 
                                         disable_output=True,mx=80, my=80, mz=80, tfinal=1.0, num_output_times=1)
        claw.run()
        test_q=claw.solution.state.get_q_global()
        nx = np.size(test_q,1)
        test_density = np.reshape(test_q[0,nx//2,:,:],np.size(test_q[0,nx//2,:,:]),order='F')
        test_err = np.linalg.norm(expected_density-test_density)
        assert test_err<1e-6
    



"""
Print status of all clawpack git repositories.
"""

from __future__ import print_function

from __future__ import absolute_import
import os
import sys
import time
import subprocess

if sys.version_info < (3, 0):
    from StringIO import StringIO
else:
    from io import StringIO

repos_list = ['', 'classic', 'amrclaw', 'clawutil', 'pyclaw', 'visclaw', 'riemann',
              'geoclaw']

def make_git_status_file(outdir=None):
    """
    Print status of all clawpack git repositories.
    Creates 2 files:
         outdir + '/claw_git_status.txt'
            contains list of last commits, what branch is checked out, and
            a short version of git status.
         outdir + '/claw_git_diffs.txt'
            contains all diffs between current working state and last commits.
        
    """

    if 'CLAW' not in os.environ:
        raise ValueError("*** CLAW environment variable not set ***")

    if outdir is None:
        outdir = os.getcwd()
    else:
        outdir = os.path.abspath(outdir)

    status_file_path = os.path.join(outdir, "claw_git_status.txt")
    diff_file_path = os.path.join(outdir, 'claw_git_diffs.txt')

    with open(status_file_path, 'w') as status_file:

        status_file.write("Clawpack Git Status \n")
        status_file.write("Diffs can be found in %s\n\n" % diff_file_path)
        status_file.write(time.strftime("%a, %d %b %Y %H:%M:%S %Z\n"))
        # status_file.write(time.strftime("%c\n"))
        status_file.write("$CLAW = %s\n" % os.environ["CLAW"])
        status_file.write("$FC = %s\n" % os.environ.get("FC", "not set"))

        for repos in repos_list:
            status_file.write(repository_status(repos))

    with open(diff_file_path, 'w') as diff_file:
        diff_file.write("Clawpack git diffs...")
        for repos in repos_list:
            diff_file.write(repository_diff(repos))


def repository_status(repository):
    r"""Return string representation of git status of *repository*

    Uses a series of system command line calls via the subprocess module.

    :Input:
     - *repository* (str) - Name of clawpack repository whose status is needed.

    :Output:
     - (str) - String with status output

    """


    # Query the repository for needed info
    repo_path = os.path.expandvars(os.path.join("$CLAW", repository.lower()))
    
    # Construct output string
    output = StringIO()

    if repository == '':
        output.write("\n\n===========\n%s\n===========\n" % 'clawpack')
    else:
        output.write("\n\n===========\n%s\n===========\n" % repository)
    output.write("%s\n\n" % repo_path)

    output.write("--- last commit ---\n")
    cmd = "cd %s ; git log -1 --oneline --decorate" % repo_path
    output.write(subprocess.check_output(cmd, shell=True,
                 universal_newlines=True))
    output.write("\n")

    output.write("--- branch and status ---\n")
    cmd = "cd %s ; git status -b -s --untracked-files=no" % repo_path
    output.write(subprocess.check_output(cmd, shell=True,
                 universal_newlines=True))
 
    output_str = output.getvalue()
    output.close()

    return output_str
    

def repository_diff(repository):
    r"""Uses a series of system command line calls via the subprocess module.

    :Input:
     - *repository* (str) - Name of clawpack repository whose status is needed.

    :Output:
     - (str) - String with diff output

    """

    # Query the repository for needed info
    repo_path = os.path.expandvars(os.path.join("$CLAW", repository.lower()))
    
    # Construct output string
    output = StringIO()

    if repository == '':
        output.write("\n\n===========\n%s\n===========\n" % 'clawpack')
    else:
        output.write("\n\n===========\n%s\n===========\n" % repository)
    output.write("%s\n\n" % repo_path)

    cmd = 'cd %s ; git diff --no-ext-diff' % repo_path
    output.write(subprocess.check_output(cmd, shell=True,
                 universal_newlines=True))
    cmd = 'cd %s ; git diff --cached --no-ext-diff' % repo_path
    output.write(subprocess.check_output(cmd, shell=True,
                 universal_newlines=True))
 
    output_str = output.getvalue()
    output.close()

    return output_str



if __name__=="__main__":
    path = os.getcwd()
    if len(sys.argv) > 1:
        path = os.path.abspath(sys.argv[1])
    make_git_status_file(path)



#!/usr/bin/env python
# encoding: utf-8

from clawpack import pyclaw
import numpy as np
import logging

def debug_loggers():
    """
    Turn on maximum debugging from all loggers.
    """

    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)
    root_logger.propagate = True


def disable_loggers():
    """
    Disable all loggers (quiet runs)
    """

    root_logger = logging.getLogger()
    root_logger.disabled = True

from clawpack.riemann import advection_1D
fsolver_1D = pyclaw.ClawSolver1D(advection_1D)
fsolver_1D.kernel_language = 'Fortran'

from clawpack.riemann import advection_1D_py
pysolver_1D = pyclaw.ClawSolver1D(advection_1D_py.advection_1D)
pysolver_1D.kernel_language = 'Python'

from clawpack.riemann import shallow_roe_with_efix_2D
fsolver_2D = pyclaw.ClawSolver2D(shallow_roe_with_efix_2D)
fsolver_2D.kernel_language = 'Fortran'

solvers_1D = {
    'current_fortran' : fsolver_1D,
    'current_python'  : pysolver_1D
}

solvers_2D = {
    'current_fortran' : fsolver_2D
}

new_solvers_1D = {}

# Here we try and bring in "experimental" solvers for comparison
# If we can't bring in the solver, complain and move on...

try:
    from clawpack.pyclaw.examples import iso_c_advection

    iso_c_solver = iso_c_advection.ISO_C_ClawSolver1D(None, 'clawpack.pyclaw')

    iso_c_solver.kernel_language = 'Fortran'
    iso_c_solver.rp = iso_c_advection.iso_c_rp1_advection(1.0)
    iso_c_solver.num_waves = 1

    new_solvers_1D['iso_c'] = iso_c_solver

except (ImportError, OSError) as err:
    print("Unable to import ISO C variant", err)

solvers_1D.update(new_solvers_1D)


def verify_1D(dx, q0, qfinal):
    if q0 is not None and qfinal is not None:
        test = dx*np.linalg.norm(qfinal-q0,1)
        return test

def verify_2D(dx, qfinal):
    if qfinal is not None:
        test = dx*np.linalg.norm(qfinal)
        return test

def compare_1D(nx=1000):
    """
    Tests a variety of Riemann solver ideas on 1D advection
    """

    from . import compare_solvers
    import time

    solvers = compare_solvers.solvers_1D

    times, tests = {}, {}

    for name, solver in solvers.items():
        solver.bc_lower[0] = pyclaw.BC.periodic
        solver.bc_upper[0] = pyclaw.BC.periodic

        from clawpack.pyclaw.examples.advection_1d import advection
        claw = advection.advection(nx=nx,outdir=compare_solvers.outdir)
        claw.solver = solver
        claw.keep_copy = True

        # benchmark
        t0 = time.clock()
        claw.run()
        t1 = time.clock()
        t = t1-t0
        times[name] = t

        # verify
        test = verify_1D(claw.solution.domain.grid.delta[0],
                         claw.frames[0].state.get_q_global(),
                         claw.frames[claw.num_output_times].state.get_q_global())
        tests[name] = test

    return times, tests


def compare_2D(nx=(250,250)):
    """
    Tests a variety of Riemann solver ideas on 2D shallow water equation
    """
    from . import compare_solvers
    import time

    solvers = compare_solvers.solvers_2D

    times, tests = {}, {}

    for name, solver in solvers.items():
        solver.num_waves = 3
        solver.bc_lower[0] = pyclaw.BC.extrap
        solver.bc_upper[0] = pyclaw.BC.wall
        solver.bc_lower[1] = pyclaw.BC.extrap
        solver.bc_upper[1] = pyclaw.BC.wall

        solver.limiters = pyclaw.limiters.tvd.MC
        solver.dimensional_split=1

        from clawpack.pyclaw.examples.shallow_2d import shallow2D
        claw = shallow2D.shallow2D(outdir=compare_solvers.outdir)
        claw.solver = solver
        claw.keep_copy = True

        t0 = time.clock()
        claw.run()
        t1 = time.clock()
        t = t1-t0
        times[name] = t

        # verify
        test = verify_2D(claw.solution.domain.grid.delta[0],
                         claw.frames[claw.num_output_times].state.get_q_global())
        tests[name] = test

    return times, tests

if __name__=="__main__":
    from . import compare_solvers
    disable_loggers()
#    debug_loggers()

    import sys

    vis = True
    
    if vis:
        compare_solvers.outdir='./_output'
    else:
        compare_solvers.outdir = None
        
    if len(sys.argv) > 1:
        nx_1D = int(sys.argv[1])
    else:
        nx_1D = 500

    if len(sys.argv) > 2:
        # '(2,2)' -> (2,2)
        nx_2D = tuple(int(i) for i in argv.split(','))
    else:
        nx_2D = 100,100

    def print_time_accuracy(times, tests, solvers):
        print("\n=====TIME=====")
        for name in solvers.keys():
            print("%-25s: %g" % (name, times[name]))

        print("\n===ACCURACY===")
        for name in solvers.keys():
            print("%-25s: %g" % (name, tests[name]))

    compare_solvers.tfinal = 1.0
    print("\nRiemann comparison on 1D advection to t=%g with %d grid points" % \
        (compare_solvers.tfinal, nx_1D))

    times, tests = compare_1D(nx=nx_1D)

    print_time_accuracy(times, tests, compare_solvers.solvers_1D)

    compare_solvers.tfinal = 2.5
    print(("\nRiemann comparison on 2D shallow water equation to t=%g" + \
               " with %dx%d grid points") % ((compare_solvers.tfinal,) + nx_2D))

    times, tests = compare_2D(nx=nx_2D)

    print_time_accuracy(times, tests, compare_solvers.solvers_2D)



#!/usr/bin/env python
# encoding: utf-8
r"""
Controller for basic computation and plotting setup.

This module defines the Pyclaw controller class.  It can be used to perform
simulations in a convenient manner similar to that available in previous
versions of Clawpack, i.e. with output_style and
output time specification.  It also can be used to set up easy plotting and 
running of compiled fortran binaries.
"""

import logging
import sys
import os
import copy

from .solver import Solver
from .util import FrameCounter
from .util import LOGGING_LEVELS

class Controller(object):
    r"""Controller for pyclaw simulation runs and plotting
            
    :Initialization:
    
        Input: None
    
    :Examples:

        >>> import clawpack.pyclaw as pyclaw
        >>> x = pyclaw.Dimension(0.,1.,100,name='x')
        >>> domain = pyclaw.Domain((x))
        >>> state = pyclaw.State(domain,3,2)
        >>> claw = pyclaw.Controller()
        >>> claw.solution = pyclaw.Solution(state,domain)
        >>> claw.solver = pyclaw.ClawSolver1D()
    """

    def __getattr__(self, key):
        if key in ('t','num_eqn','mp','mF','q','p','F','aux','capa',
                   'problem_data','num_aux',
                   'num_dim', 'p_centers', 'p_edges', 'c_centers', 'c_edges',
                   'num_cells', 'lower', 'upper', 'delta', 'centers', 'edges',
                   'gauges', 'num_eqn', 'num_aux', 'grid', 'problem_data'):
            return self._get_solution_attribute(key)
        else:
            raise AttributeError("'Controller' object has no attribute '"+key+"'")

    def _get_solution_attribute(self, name):
        r"""
        Return solution attribute
        
        :Output:
         - (id) - Value of attribute from ``solution``
        """
        return getattr(self.solution,name)
 

    #  ======================================================================
    #   Property Definitions
    #  ======================================================================
    @property
    def verbosity(self):
        return self._verbosity

    @verbosity.setter
    def verbosity(self, value):
        self._verbosity = value
        # Only adjust console logger; leave file logger alone
        self.logger.handlers[1].setLevel(LOGGING_LEVELS[value])

    @property
    def outdir_p(self):
        r"""(string) - Directory to use for writing derived quantity files"""
        return os.path.join(self.outdir,'_p')
    @property
    def F_path(self):
        r"""(string) - Full path to output file for functionals"""
        return os.path.join(self.outdir,self.F_file_name+'.txt')

    #  ======================================================================
    #   Initialization routines
    #  ======================================================================
    def __init__(self):
        r"""
        Initialization routine for a Controller object.
        
        See :class:`Controller` for full documentation.
        """
        
        import numpy as np

        self.viewable_attributes = ['xdir','rundir','outdir','overwrite',
                        'xclawcmd','xclawout','xclawerr','runmake','savecode',
                        'solver','keep_copy','write_aux_init',
                        'write_aux_always','output_format',
                        'output_file_prefix','output_options','num_output_times',
                        'output_style','verbosity']
        r"""(list) - Viewable attributes of the `:class:`~pyclaw.controller.Controller`"""

        # Global information for running and/or plotting
        self.xdir = os.getcwd()
        r"""(string) - Executable path, executes xclawcmd in xdir"""
        self.rundir = os.getcwd()
        r"""(string) - Directory to run from (containing \*.data files), uses 
        \*.data from rundir"""
        self.outdir = os.getcwd()+'/_output'
        r"""(string) - Output directory, directs output files to outdir"""
        self.overwrite = True
        r"""(bool) - Ok to overwrite old result in outdir, ``default = True``"""

        self.xclawcmd = 'xclaw'
        r"""(string) - Command to execute (if using fortran), defaults to xclaw or
        xclaw.exe if cygwin is being used (which it checks vis sys.platform)"""
        if sys.platform == 'cygwin':
            self.xclawcmd = 'xclaw.exe'

        self.start_frame = 0
        self.xclawout = None
        r"""(string) - Where to write timestep messages"""
        self.xclawerr = None
        r"""(string) - Where to write error messages"""
        self.runmake = False
        r"""(bool) - Run make in xdir before xclawcmd"""
        self.savecode = False
        r"""(bool) - Save a copy of \*.f files in outdir"""
        
        self.setplot = None

        # Solver information
        self.solution = None
        self.solver = None
        r"""(:class:`~pyclaw.solver.Solver`) - Solver object"""
        
        # Output parameters for run convenience method
        self.keep_copy = False 
        r"""(bool) - Keep a copy in memory of every output time, 
        ``default = False``"""
        self.frames = []
        r"""(list) - List of saved frames if ``keep_copy`` is set to ``True``"""
        self.write_aux_init = False
        r"""(bool) - Write out initial auxiliary array, ``default = False``"""
        self.write_aux_always = False
        r"""(bool) - Write out auxiliary array at every time step, 
        ``default = False``"""
        self.output_format = 'ascii'
        r"""(list of strings) - Format or list of formats to output the data, 
        if this is None, no output is performed.  See _pyclaw_io for more info
        on available formats.  ``default = 'ascii'``"""
        self.output_file_prefix = None
        r"""(string) - File prefix to be appended to output files, 
        ``default = None``"""
        self.output_options = {}
        r"""(dict) - Output options passed to function writing and reading 
        data in output_format's format.  ``default = {}``"""
        
        self.logger = logging.getLogger('pyclaw.controller')

        # Classic output parameters, used in run convenience method
        self.tfinal = 1.0
        r"""(float) - Final time output, ``default = 1.0``"""
        self.output_style = 1
        r"""(int) - Time output style, ``default = 1``"""
        self.verbosity = 3
        r"""(int) - Level of output to screen; ``default = 3``"""
        self.num_output_times = 10                  # Outstyle 1 defaults
        r"""(int) - Number of output times, only used with ``output_style = 1``,
        ``default = 10``"""
        self.out_times = np.linspace(0.0,self.tfinal,self.num_output_times-
                                     self.start_frame)  # Outstyle 2
        r"""(list of floats) - Output time list, only used with ``output_style = 2``,
        ``default = numpy.linspace(0.0,tfinal,num_output_times)``"""

        self.nstepout = 1               # Outstyle 3 defaults
        r"""(int) - Number of steps between output, only used with
        ``output_style = 3``, ``default = 1``"""

        # Data objects
        self.plotdata = None
        r"""(:class:`~visclaw.data.ClawPlotData`) - An instance of a
        :class:`~visclaw.data.ClawPlotData` object defining the
        objects plot parameters."""

        # Derived quantity p
        self.file_prefix_p = 'claw_p'
        r"""(string) - File prefix to be prepended to derived quantity output files"""
        self.compute_p = None
        r"""(function) - function that computes derived quantities"""

        # functionals
        self.compute_F = None
        r"""(function) - Function that computes density of functional F"""
        self.F_file_name = 'F'
        r"""(string) - Name of text file containing functionals"""

    # ========== Access methods ===============================================
    def __str__(self):
        output = "Controller attributes:\n"
        for attr in self.viewable_attributes:
            value = getattr(self,attr)
            output = output + "  %s = %s \n" % (attr,value)
        output = output + '\n'
        if self.plotdata is not None:
            output = output + "  Data "+str(self.plotdata)+"\n"
        if self.solver is not None:
            output = output + "  Solver "+str(self.solver)+"\n"
        if len(self.frames) > 0:
            output = output + "  Frames \n"
            for frame in self.frames:
                output = output + "    " + str(frame) + "\n"
        return output

    # ========== Properties ==================================================

    def check_validity(self):
        r"""Check that the controller has been properly set up and is ready to run.

            Also checks validity of the solver, solution and states.
        """
        # Check to make sure we have a valid solver to use
        if self.solver is None:
            raise Exception("No solver set in controller.")
        if not isinstance(self.solver,Solver):
            raise Exception("Solver is not of correct type.")
        valid, reason = self.solver.is_valid()
        if not valid:
            raise Exception("The solver failed to initialize properly because "+reason)

        # Check to make sure the initial solution is valid
        if not self.solution.is_valid():
            raise Exception("Initial solution is not valid.")
        if not all([state.is_valid() for state in self.solution.states]):
            raise Exception("Initial states are not valid.")

    # ========== Plotting methods ============================================
    def set_plotdata(self):
        from clawpack.visclaw import data
        from clawpack.visclaw import frametools
        plotdata = data.ClawPlotData()
        plotdata.setplot = self.setplot
        self.plotdata = frametools.call_setplot(self.setplot,plotdata)
        plotdata._mode = 'iplotclaw'

    def load_frame(self,frame_number):
        try:
            return self.frames[frame_number]
        except IndexError:
            print("Cannot plot frame %s; only %s frames available" % (frame_number, len(self.frames)))

    def plot_frame(self, frame):
        if self.plotdata is None:
            self.set_plotdata()

        if frame is not None:
            frameno = self.frames.index(frame)
            from clawpack.visclaw import frametools
            frametools.plot_frame(frame, self.plotdata, frameno=frameno)

    def plot(self):
        """Plot from memory."""
        if len(self.frames) == 0:  # No frames to plot
            print("No frames to plot.  Did you forget to run, or to set keep_copy=True?")
            return

        from clawpack.visclaw import iplot

        if self.plotdata is None:
            self.set_plotdata()

        ip = iplot.Iplot(self.load_frame,self.plot_frame)
        ip.plotloop()

    # ========== Solver convenience methods ==================================
    def run(self):
        r"""
        Convenience routine that will evolve solution based on the
        traditional clawpack output and run parameters.

        This function uses the run parameters and solver parameters to evolve
        the solution to the end time specified in run_data, outputting at the
        appropriate times.

        :Input:
            None

        :Output:
            (dict) - Return a dictionary of the status of the solver.
        """
        import numpy as np

        if self.solver is None or self.solution is None:
            raise Exception('To run, a Controller must have a Solver and a Solution.')

        self.start_frame = self.solution.start_frame
        if len(self.solution.patch.grid.gauges)>0:
            self.solution.patch.grid.setup_gauge_files(self.outdir)
        frame = FrameCounter()

        frame.set_counter(self.start_frame)

        if not self.solver._is_set_up:
            self.solver.setup(self.solution)
            self.solver.dt = self.solver.dt_initial

        self.check_validity()

        # Write initial gauge values
        self.solver.write_gauge_values(self.solution)

        # Output styles
        if self.output_style == 1:
            output_times = np.linspace(self.solution.t,self.tfinal,
                                       self.num_output_times+1)
        elif self.output_style == 2:
            output_times = self.out_times
        elif self.output_style == 3:
            output_times = np.ones((self.num_output_times+1-
                                    self.start_frame))
        else:
            raise Exception("Invalid output style %s" % self.output_style)

        if len(output_times) == 0:
            print("No valid output times; halting.")
            if self.t == self.tfinal:
                print("Simulation has already reached tfinal.")
            return None

        # Output and save initial frame
        if self.keep_copy:
            self.frames.append(copy.deepcopy(self.solution))
        if self.output_format is not None:
            if os.path.exists(self.outdir) and self.overwrite is False:
                raise Exception("Refusing to overwrite existing output data. \
                 \nEither delete/move the directory or set controller.overwrite=True.")
            if self.compute_p is not None:
                self.compute_p(self.solution.state)
                self.solution.write(frame,self.outdir_p,
                                        self.output_format,
                                        self.file_prefix_p,
                                        write_aux=False,
                                        options=self.output_options,
                                        write_p=True)

            write_aux = (self.write_aux_always or self.write_aux_init)
            self.solution.write(frame,self.outdir,
                                        self.output_format,
                                        self.output_file_prefix,
                                        write_aux,
                                        self.output_options)

        self.write_F('w')

        self.log_info("Solution %s computed for time t=%f" %
                        (frame,self.solution.t) )

        for t in output_times[1:]:
            if self.output_style < 3:
                status = self.solver.evolve_to_time(self.solution,t)
            else:
                # Take nstepout steps and output
                for n in range(self.nstepout):
                    status = self.solver.evolve_to_time(self.solution)
            frame.increment()
            if self.keep_copy:
                # Save current solution to dictionary with frame as key
                self.frames.append(copy.deepcopy(self.solution))
            if self.output_format is not None:
                if self.compute_p is not None:
                    self.compute_p(self.solution.state)
                    self.solution.write(frame,self.outdir_p,
                                            self.output_format,
                                            self.file_prefix_p,
                                            write_aux=False,
                                            options=self.output_options,
                                            write_p=True)

                self.solution.write(frame,self.outdir,
                                            self.output_format,
                                            self.output_file_prefix,
                                            self.write_aux_always,
                                            self.output_options)
            self.write_F()

            self.log_info("Solution %s computed for time t=%f"
                          % (frame,self.solution.t))
            for gfile in self.solution.state.grid.gauge_files:
                gfile.flush()

        for gfile in self.solution.state.grid.gauge_files: gfile.close()

        self.solution._start_frame = len(self.frames)

        # Return the current status of the solver
        return status

    # ========== Advanced output methods ==================================

    def write_F(self,mode='a'):
        if self.compute_F is not None:
            self.compute_F(self.solution.state)
            F = [0]*self.solution.state.mF
            for i in range(self.solution.state.mF):
                F[i] = self.solution.state.sum_F(i)
            if self.is_proc_0():
                t=self.solution.t
                F_file = open(self.F_path,mode)
                F_file.write(str(t)+' '+' '.join(str(j) for j in F) + '\n')
                F_file.close()

    def is_proc_0(self):
        return True

    def log_info(self, str):
        self.logger.info(str)




class OutputController(object):
    r"""Output controller object providing an interface to the io package.

    This controller is meant to provide a uniform interface to the IO
    capabilities in PyClaw.

    :Note: This object is meant to be a test bed for restructuring the way
    PyClaw does IO and should be used at one's own risk as rapid changes could
    occur that will break backwards compatibility.

    :Initialization:

        Input:
            - *output_path* - (path) Path to output both for reading and writing.
            - *file_format* - (string) String indicating type of file format
              to be used.  Default is `ascii`.

    .. attribute:: output_path

        Path to output directory either for writing or reading.

    .. attribute:: file_format

        Format used for reading and writing.  Should be one of the modules in
        the io package.  Changing this attribute will automatically load the
        new backend io package.

    """

    @property
    def file_format(self):
        return self._file_format
    @file_format.setter
    def file_format(self, value):
        if value.lower() == 'petsc':
            self._io_module = __import__("clawpack.petclaw.fileio.petsc",
                                         fromlist=["clawpack.petclaw.fileio"])
            self._file_format = value
        elif value.lower()[:6] == 'binary':
            # could be 'binary64' or 'binary32'
            self._io_module = __import__("clawpack.pyclaw.fileio.binary",
                                         fromlist=['clawpack.pyclaw.fileio'])
            self._file_format = value
        else:
            self._io_module = __import__("clawpack.pyclaw.fileio.%s" % value,
                                         fromlist=['clawpack.pyclaw.fileio'])
            self._file_format = value



    def __init__(self, output_path, file_format='ascii', file_prefix='fort'):
        r"""
        Initialization routine for an OutputController object.

        See :class:`OutputController` for full documentation.
        """

        self._io_module = None
        self.file_format = file_format
        self.output_path = output_path
        self.file_prefix = file_prefix


    def get_time(self, frame_num):
        r"""Get the time for the specified frame number.

        :Input:
            - *frame_num* - (int) Frame number of solution to fetch.

        :Output:
            - (float or None) Returns either the time from the specified
              *frame_num* if the io format allows for it or returns None.

        """

        if "read_t" in dir(self._io_module):
            return self._io_module.read_t(frame_num, path=self.output_path,
                                          file_prefix=self.file_prefix)[0]
        else:
            return None



if __name__ == "__main__":
    import doctest
    doctest.testmod()




from __future__ import absolute_import
from pylab import *

outdir = '_output'
d = loadtxt(outdir + '/fort.FG1.valuemax')
b = loadtxt(outdir + '/fort.FG1.aux1')
topo = b[:,4]  # for level 3
figure(401)
clf()
